/**
 * Bull SAS / OW2 Orchestra
 * Copyright (C) 2010 Bull S.A.S, and individual contributors as indicated
 * by the @authors tag.
 *
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.designer.models.transformers.bpmn2designer;

import java.util.ArrayList;
import java.util.List;
import java.util.ListIterator;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import javax.xml.XMLConstants;
import javax.xml.bind.JAXBElement;
import javax.xml.namespace.QName;
import org.ow2.orchestra.designer.bpmn.model.data.type.WSDLMessageModel;
import org.ow2.orchestra.designer.bpmn.model.data.type.WSDLMessagePartModel;
import org.ow2.orchestra.designer.bpmn.model.operations.CorrelationMappingModel;
import org.ow2.orchestra.designer.bpmn.model.operations.CorrelationModel;
import org.ow2.orchestra.designer.models.exceptions.TransformationException;
import org.ow2.orchestra.jaxb.bpmn.CorrelationKey;
import org.ow2.orchestra.jaxb.bpmn.CorrelationPropertyRetrievalExpression;
import org.ow2.orchestra.jaxb.bpmn.MessageFlow;
import org.ow2.orchestra.jaxb.bpmn.TCollaboration;
import org.ow2.orchestra.jaxb.bpmn.TConversationNode;
import org.ow2.orchestra.jaxb.bpmn.TCorrelationProperty;
import org.ow2.orchestra.jaxb.bpmn.TFormalExpression;
import org.ow2.orchestra.jaxb.bpmn.TRootElement;

/**
 * Transform {@link org.ow2.orchestra.jaxb.bpmn.TCorrelationProperty}s into {@link org.ow2.orchestra.designer.bpmn.model.operations.CorrelationModel}s
 *
 * @author Guillaume Porcher
 */
public final class CorrelationTransformer {

  /**
   * Utility classes should not have public or default constructors.
   */
  private CorrelationTransformer() {

  }

  /**
   * @throws org.ow2.orchestra.designer.models.exceptions.TransformationException
   *
   */
  public static CorrelationModel bpmnToModelCorrelation(
          final TCorrelationProperty correlationProperty,
          final Bpmn2DesignerContext context) throws TransformationException {
    if (!context.getCorrelationsMap().containsKey(correlationProperty)) {
      CorrelationModel correlationModel = new CorrelationModel();
      correlationModel.setName(correlationProperty.getName());
      correlationModel.setId(correlationProperty.getId());
      correlationModel.setType(ElementTransformer.getDataTypeModel(context, correlationProperty.getType().getLocalPart()));

      context.getProcessModel().getCorrelations().add(correlationModel);
      context.getCorrelationsMap().put(correlationProperty, correlationModel);
    }
    return context.getCorrelationsMap().get(correlationProperty);
  }

  private static TCorrelationProperty getCorrelationProperty(String correlationId, final Bpmn2DesignerContext context) {
    for (final JAXBElement<? extends TRootElement> rootElement : context.getDefinitions().getRootElements()) {
      if (rootElement.getValue() instanceof TCorrelationProperty) {
        if (rootElement.getValue().getId().equals(correlationId)) {
          return (TCorrelationProperty) rootElement.getValue();
        }
      }
    }
    return null;
  }

  public static List<CorrelationMappingModel> bpmnToModelCorrelationMapping(final MessageFlow messageFlow, final Bpmn2DesignerContext context) throws TransformationException {
    List<CorrelationMappingModel> result = new ArrayList<CorrelationMappingModel>();
    for (final JAXBElement<? extends TRootElement> rootElement : context.getDefinitions().getRootElements()) {
      if (rootElement.getDeclaredType().equals(TCollaboration.class)) {
        final TCollaboration collaboration = (TCollaboration) rootElement.getValue();
        for (final JAXBElement<? extends TConversationNode> conversationElement : collaboration.getConversationNodes()) {
          TConversationNode conversationNode = conversationElement.getValue();
          if (!conversationNode.getCorrelationKeies().isEmpty()
            && conversationNode.getMessageFlowReves().contains(new QName(context.getDefinitions().getTargetNamespace(), messageFlow.getId()))) {
            // Add correlation mapping model
            for (CorrelationKey correlationKey : conversationNode.getCorrelationKeies()) {
              for (QName correlationPropertyQName : correlationKey.getCorrelationPropertyReves()) {
                CorrelationMappingModel correlationMappingModel = new CorrelationMappingModel();
                correlationMappingModel.setMessageModel((WSDLMessageModel) ElementTransformer.getDataTypeModel(context, context.getMessage(messageFlow.getMessageRef()).getItemRef().getLocalPart()));
                TCorrelationProperty correlationProperty = getCorrelationProperty(correlationPropertyQName.getLocalPart(), context); // TODO: handle multiple namespaces
                correlationMappingModel.setCorrelationModel(bpmnToModelCorrelation(correlationProperty, context));
                for (CorrelationPropertyRetrievalExpression correlationPropertyRetrievalExpression : correlationProperty.getCorrelationPropertyRetrievalExpressions()) {
                  if (correlationPropertyRetrievalExpression.getMessageRef().equals(messageFlow.getMessageRef())) {
                    TFormalExpression formalExpression = correlationPropertyRetrievalExpression.getMessagePath();
                    List<Object> queryExpressionObjects = new ArrayList<Object>(formalExpression.getContent());
                    final ListIterator<Object> iterator = queryExpressionObjects.listIterator();
                    String queryExpression = "";
                    while (iterator.hasNext()) {
                      final Object o = iterator.next();
                      if (o instanceof String) {
                        queryExpression = (String) o;
                      }
                    }
                    Matcher matcher = Pattern.compile("^/(.*)/(.*)$").matcher(queryExpression);
                    String partName = null;
                    String selector = null;
                    if (matcher.find()) {
                      partName = matcher.group(1);
                      selector = matcher.group(2);
                    } else {
                      Matcher matcher2 = Pattern.compile("^/(.*)$").matcher(queryExpression);
                      if (matcher2.find()) {
                        partName = matcher2.group(1);
                      } else {
                        throw new TransformationException("message part not found in correlation property retrieval expression " + queryExpression);
                      }
                    }
                    for (WSDLMessagePartModel partModel: correlationMappingModel.getMessageModel().getParts()) {
                      if (partModel.getPartName().equals(partName)) {
                        correlationMappingModel.setMessagePartModel(partModel);
                      }
                    }
                    if (selector != null) {
                      String[] tmp = selector.split(":");
                      if (tmp.length == 2) {
                        String prefix = tmp[0];
                        String namespace = formalExpression.getOtherAttributes().get(new QName(XMLConstants.XMLNS_ATTRIBUTE_NS_URI, prefix));
                        selector = "{" + namespace + "}" + tmp[1];
                      }
                    }
                    correlationMappingModel.setChildNodeSelector(selector);
                  }
                }

                result.add(correlationMappingModel);
              }
            }
          }
        }
      }
    }
    return result;
  }
}
