/**
 * Bull SAS / OW2 Orchestra
 * Copyright (C) 2010 Bull S.A.S, and individual contributors as indicated
 * by the @authors tag.
 *
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.designer.models.transformers.bpmn2designer;

import java.util.ArrayList;
import java.util.List;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import javax.xml.XMLConstants;
import javax.xml.bind.JAXBElement;
import javax.xml.namespace.QName;
import org.ow2.orchestra.designer.bpmn.model.AbstractElement;
import org.ow2.orchestra.designer.bpmn.model.AbstractElementWithPosition;
import org.ow2.orchestra.designer.bpmn.model.FlowModel;
import org.ow2.orchestra.designer.bpmn.model.data.DataInputModel;
import org.ow2.orchestra.designer.bpmn.model.data.DataMappingModel;
import org.ow2.orchestra.designer.bpmn.model.data.DataObjectModel;
import org.ow2.orchestra.designer.bpmn.model.data.ExpressionDataInputModel;
import org.ow2.orchestra.designer.bpmn.model.data.ExpressionDataObjectModel;
import org.ow2.orchestra.designer.bpmn.model.data.ExpressionOperationModel;
import org.ow2.orchestra.designer.bpmn.model.data.LiteralModel;
import org.ow2.orchestra.designer.bpmn.model.data.OperationType;
import org.ow2.orchestra.designer.bpmn.model.data.type.DataTypeModel;
import org.ow2.orchestra.designer.bpmn.model.data.type.XSDSimpleType;
import org.ow2.orchestra.designer.bpmn.model.process.AbstractBPMNElementWithData;
import org.ow2.orchestra.designer.bpmn.model.process.IBPMNElementWithData;
import org.ow2.orchestra.designer.models.exceptions.TransformationException;
import org.ow2.orchestra.jaxb.bpmn.Assignment;
import org.ow2.orchestra.jaxb.bpmn.DataInput;
import org.ow2.orchestra.jaxb.bpmn.DataInputAssociation;
import org.ow2.orchestra.jaxb.bpmn.TDataObject;
import org.ow2.orchestra.jaxb.bpmn.TExpression;
import org.ow2.orchestra.jaxb.bpmn.TFlowElement;
import org.ow2.orchestra.jaxb.bpmn.TProcess;
import org.ow2.orchestra.jaxb.bpmn.TRootElement;

/**
 * Transform {@link DataInputAssociation}s into {@link org.ow2.orchestra.designer.bpmn.model.data.DataMappingModel}s
 *
 * @author Guillaume Porcher
 */
public final class DataMapperTransformer {

  /**
   * Utility classes should not have public or default constructors.
   */
  private DataMapperTransformer() {

  }

  /**
   * @throws org.ow2.orchestra.designer.models.exceptions.TransformationException
   *
   */
  private static void bpmnToModel(
          final DataInputAssociation dataInputAssociation,
          final DataMappingModel dataMappingModel,
          final IBPMNElementWithData elementWithData,
          final Bpmn2DesignerContext context,
          final int position) throws TransformationException {
    for (Assignment assignment : dataInputAssociation.getAssignments()) {
      // Parse to expr
      ExpressionDataInputModel expressionDataInputModel =
              getToExpressionModel(assignment.getTo(), dataMappingModel, elementWithData, position);
      // Parse from expr
      AbstractElementWithPosition fromExpressionModel =
              getFromExpressionModel(assignment.getFrom(), dataMappingModel, context, position);
      if (fromExpressionModel == null) {
        throw new TransformationException("Unable to generate expression.");
      }
      // Create link between from expr and to expr
      FlowModel flowModel = new FlowModel();
      flowModel.setSourceElement(fromExpressionModel);
      fromExpressionModel.getOutgoingFlows().add(flowModel);
      flowModel.setTargetElement(expressionDataInputModel);
      expressionDataInputModel.getIncomingFlows().add(flowModel);
      dataMappingModel.getElements().add(flowModel);
    }
  }

  private static ExpressionDataInputModel getToExpressionModel(final TExpression toExpression,
                                                               final DataMappingModel dataMappingModel,
                                                               final IBPMNElementWithData elementWithData,
                                                               final int position) {
    // Parse to expr
    String toExpr = "";
    for (final Object content : toExpression.getContent()) {
      if (content instanceof String && !"".equals(content)) {
        toExpr = (String) content;
      }
    }
    ExpressionDataInputModel expressionDataInputModel = new ExpressionDataInputModel();

    DataInputModel dataInputModel = null;
    String childNodeSelector = null;
    // Compile and use regular expression
    String dataInputPatternStr = "getDataInput\\(\"(.*)\"\\)/(.*)";
    Pattern pattern = Pattern.compile(dataInputPatternStr);
    Matcher matcher = pattern.matcher(toExpr);
    boolean matchFound = matcher.find();
    if (matchFound) {
      String dataInputName = matcher.group(1);
      String childNS = matcher.group(2);
      String[] tmp = childNS.split(":");
      if (tmp.length == 2) {
        String prefix = tmp[0];
        String namespace = toExpression.getOtherAttributes().get(new QName(XMLConstants.XMLNS_ATTRIBUTE_NS_URI, prefix));
        childNodeSelector = "{" + namespace + "}" + tmp[1];
      } else {
        childNodeSelector = childNS;
      }
      for (DataInputModel dim : elementWithData.getDataInputs()) {
        if (dim.getName().equals(dataInputName)) {
          dataInputModel = dim;
        }
      }
    } else {
      String dataInputPatternStr2 = "getDataInput\\(\"(.*)\"\\)";
      Pattern pattern2 = Pattern.compile(dataInputPatternStr2);
      Matcher matcher2 = pattern2.matcher(toExpr);
      boolean matchFound2 = matcher2.find();
      if (matchFound2) {
        String dataInputName = matcher2.group(1);
        for (DataInputModel dim : elementWithData.getDataInputs()) {
          if (dim.getName().equals(dataInputName)) {
            dataInputModel = dim;
          }
        }
      }
    }
    // set data input
    expressionDataInputModel.setDataInputModel(dataInputModel);
    expressionDataInputModel.setChildNodeSelector(childNodeSelector);
    dataMappingModel.getElements().add(expressionDataInputModel);
    return expressionDataInputModel;
  }

  private static AbstractElementWithPosition parseGetDataObject(
          final String fromExpr,
          final TExpression fromExpression,
          final DataMappingModel dataMappingModel,
          final Bpmn2DesignerContext context,
          final int position) throws TransformationException {
    DataObjectModel dataObjectModel = null;
    String childNodeSelector = null;
    // Compile and use regular expression for DataObject
    String dataObjectPatternStr = "^getDataObject\\(\"(.*)\"\\)/(.*)$";
    Pattern pattern = Pattern.compile(dataObjectPatternStr);
    Matcher matcher = pattern.matcher(fromExpr);
    boolean matchFound = matcher.find();
    if (matchFound) {
      String dataObjectName = matcher.group(1);
      String childNS = matcher.group(2);
      String[] tmp = childNS.split(":");
      if (tmp.length == 2) {
        String prefix = tmp[0];
        String namespace = fromExpression.getOtherAttributes().get(new QName(XMLConstants.XMLNS_ATTRIBUTE_NS_URI, prefix));
        childNodeSelector = "{" + namespace + "}" + tmp[1];
      } else {
        childNodeSelector = childNS;
      }
      // get DataObjectModel from Process
      dataObjectModel = getDataObject(context, dataObjectName);
      if (dataObjectModel == null) {
        throw new TransformationException("No dataObject '" + dataObjectName + "' found in expression " + fromExpr);
      }
      ExpressionDataObjectModel expressionDataObjectModel = new ExpressionDataObjectModel();

      expressionDataObjectModel.setDataObjectModel(dataObjectModel);
      expressionDataObjectModel.setChildNodeSelector(childNodeSelector);
      dataMappingModel.getElements().add(expressionDataObjectModel);
      return expressionDataObjectModel;
    }
    String dataObjectPatternStr2 = "^getDataObject\\(\"(.*)\"\\)$";
    Pattern pattern2 = Pattern.compile(dataObjectPatternStr2);
    Matcher matcher2 = pattern2.matcher(fromExpr);
    boolean matchFound2 = matcher2.find();
    if (matchFound2) {
      String dataObjectName = matcher2.group(1);
      // get DataObjectModel from Process
      dataObjectModel = getDataObject(context, dataObjectName);
      if (dataObjectModel == null) {
        throw new TransformationException("No dataObject '" + dataObjectName + "' found in expression " + fromExpr);
      }
      ExpressionDataObjectModel expressionDataObjectModel = new ExpressionDataObjectModel();

      expressionDataObjectModel.setDataObjectModel(dataObjectModel);
      expressionDataObjectModel.setChildNodeSelector(childNodeSelector);
      dataMappingModel.getElements().add(expressionDataObjectModel);
      return expressionDataObjectModel;
    }
    return null;
  }

  private static DataObjectModel getDataObject(final Bpmn2DesignerContext context, final String dataObjectName) throws TransformationException {
    for (final JAXBElement<? extends TRootElement> rootElement : context.getDefinitions().getRootElements()) {
      if (rootElement.getDeclaredType().equals(TProcess.class)) {
        final TProcess tProcess = (TProcess) rootElement.getValue();
        for (final JAXBElement<? extends TFlowElement> flowElement : tProcess.getFlowElements()) {
          if (flowElement.getValue() instanceof TDataObject) {
            TDataObject dataObject = (TDataObject) flowElement.getValue();
            if (dataObject.getName().equals(dataObjectName)) {
              ElementTransformer.bpmnToModel(dataObject, context);
              return (DataObjectModel) context.getModelsMap().get(dataObject);
            }
          }
        }
      }
    }
    return null;
  }

  private static AbstractElementWithPosition parseLiteral(
          final String fromExpr,
          final TExpression fromExpression,
          final DataMappingModel dataMappingModel,
          final Bpmn2DesignerContext context,
          final int position) throws TransformationException {
    LiteralModel literalModel = new LiteralModel();
    XSDSimpleType literalType = new XSDSimpleType();
    literalModel.setType(literalType);
    dataMappingModel.getElements().add(literalModel);

    Matcher literalDateMatcher = Pattern.compile("^\"(\\d{4}-\\d{2}-\\d{2}T\\d{2}:\\d{2}:\\d{2})\"$").matcher(fromExpr);
    if (literalDateMatcher.find()) {
      String literalValue = literalDateMatcher.group(1);
      literalType.setQName(new QName(XMLConstants.W3C_XML_SCHEMA_NS_URI, "dateTime"));
      literalModel.setValue(literalValue);
      return literalModel;
    }

    Matcher literalStringMatcher = Pattern.compile("^\"(.*)\"$").matcher(fromExpr);
    if (literalStringMatcher.find()) {
      String literalValue = literalStringMatcher.group(1);
      literalType.setQName(new QName(XMLConstants.W3C_XML_SCHEMA_NS_URI, "string"));
      literalModel.setValue(literalValue);
      return literalModel;
    }

    Matcher literalNumberMatcher = Pattern.compile("^(-?\\d+\\.?\\d*|-?\\.\\d+)$").matcher(fromExpr);
    if (literalNumberMatcher.find()) {
      String literalValue = literalNumberMatcher.group(1);
      literalType.setQName(new QName(XMLConstants.W3C_XML_SCHEMA_NS_URI, "decimal"));
      literalModel.setValue(literalValue);
      return literalModel;
    }

    Matcher literalTrueMatcher = Pattern.compile("^true\\(\\)$").matcher(fromExpr);
    if (literalTrueMatcher.find()) {
      literalType.setQName(new QName(XMLConstants.W3C_XML_SCHEMA_NS_URI, "boolean"));
      literalModel.setValue("true");
      return literalModel;
    }

    Matcher literalFalseMatcher = Pattern.compile("^false\\(\\)$").matcher(fromExpr);
    if (literalFalseMatcher.find()) {
      literalType.setQName(new QName(XMLConstants.W3C_XML_SCHEMA_NS_URI, "boolean"));
      literalModel.setValue("false");
      return literalModel;
    }

    return null;
  }

  private static int countStringOccurrences(final CharSequence input, final String pattern) {
    int result = 0;
    Matcher stringMatcher = Pattern.compile(pattern).matcher(input);
    while (stringMatcher.find()) {
      result++;
    }
    return result;
  }

  private static boolean areOperandsValid(final String op1Value, final String op2Value) {
    return (!op1Value.contains("\"") || countStringOccurrences(op1Value, "\"") % 2 == 0) &&
           (!op2Value.contains("\"") || countStringOccurrences(op1Value, "\"") % 2 == 0) &&
           (countStringOccurrences(op1Value, "\\(") == countStringOccurrences(op1Value, "\\)")) &&
           (countStringOccurrences(op2Value, "\\(") == countStringOccurrences(op2Value, "\\)"));
  }

  private static AbstractElementWithPosition parseOperation(
          final String fromExpr,
          final TExpression fromExpression,
          final DataMappingModel dataMappingModel,
          final Bpmn2DesignerContext context,
          final int position) throws TransformationException {
    String op1Value = null;
    String op2Value = null;
    OperationType operationType = null;

    // Parse concat
    Matcher concatMatcher = Pattern.compile("^concat\\((.+),(.+)\\)$").matcher(fromExpr);
    if (operationType == null && concatMatcher.find()) {
      op1Value = concatMatcher.group(1);
      op2Value = concatMatcher.group(2);
      //Check that if op1 and op2 contains quote chars then it should contains a pair number of those chars
      if (areOperandsValid(op1Value, op2Value)) {
        operationType = OperationType.CONCATENATE;
      }
    }
    // parse +
    Matcher addMatcher = Pattern.compile("^(.+)\\+(.+)$").matcher(fromExpr);
    if (operationType == null && addMatcher.find()) {
      op1Value = addMatcher.group(1);
      op2Value = addMatcher.group(2);
      //Check that if op1 and op2 contains quote chars then it should contains a pair number of those chars
      if (areOperandsValid(op1Value, op2Value)) {
        operationType = OperationType.ADD;
      }
    }
    // parse div
    Matcher divMatcher = Pattern.compile("^(.+) div (.+)$").matcher(fromExpr);
    if (operationType == null && divMatcher.find()) {
      op1Value = divMatcher.group(1);
      op2Value = divMatcher.group(2);
      //Check that if op1 and op2 contains quote chars then it should contains a pair number of those chars
      if (areOperandsValid(op1Value, op2Value)) {
        operationType = OperationType.DIVIDE;
      }
    }
    // parse *
    Matcher multMatcher = Pattern.compile("^(.+)\\*(.+)$").matcher(fromExpr);
    if (operationType == null && multMatcher.find()) {
      op1Value = multMatcher.group(1);
      op2Value = multMatcher.group(2);
      //Check that if op1 and op2 contains quote chars then it should contains a pair number of those chars
      if (areOperandsValid(op1Value, op2Value)) {
        operationType = OperationType.MULTIPLY;
      }
    }
    // parse >=
    Matcher getMatcher = Pattern.compile("^(.+)>=(.+)$").matcher(fromExpr);
    if (operationType == null && getMatcher.find()) {
      op1Value = getMatcher.group(1);
      op2Value = getMatcher.group(2);
      //Check that if op1 and op2 contains quote chars then it should contains a pair number of those chars
      if (areOperandsValid(op1Value, op2Value)) {
        operationType = OperationType.IS_GREATER_OR_EQUALS;
      }
    }
    // parse >
    Matcher gtMatcher = Pattern.compile("^(.+)>(.+)$").matcher(fromExpr);
    if (operationType == null && gtMatcher.find()) {
      op1Value = gtMatcher.group(1);
      op2Value = gtMatcher.group(2);
      //Check that if op1 and op2 contains quote chars then it should contains a pair number of those chars
      if (areOperandsValid(op1Value, op2Value)) {
        operationType = OperationType.IS_GREATER_THAN;
      }
    }

    // parse <=
    Matcher letMatcher = Pattern.compile("^(.+)<=(.+)$").matcher(fromExpr);
    if (operationType == null && letMatcher.find()) {
      op1Value = letMatcher.group(1);
      op2Value = letMatcher.group(2);
      //Check that if op1 and op2 contains quote chars then it should contains a pair number of those chars
      if (areOperandsValid(op1Value, op2Value)) {
        operationType = OperationType.IS_LESS_OR_EQUALS;
      }
    }
    // parse >
    Matcher ltMatcher = Pattern.compile("^(.+)<(.+)$").matcher(fromExpr);
    if (operationType == null && ltMatcher.find()) {
      op1Value = ltMatcher.group(1);
      op2Value = ltMatcher.group(2);
      //Check that if op1 and op2 contains quote chars then it should contains a pair number of those chars
      if (areOperandsValid(op1Value, op2Value)) {
        operationType = OperationType.IS_LESS_THAN;
      }
    }
    // parse !=
    Matcher neMatcher = Pattern.compile("^(.+)!=(.+)$").matcher(fromExpr);
    if (operationType == null && neMatcher.find()) {
      op1Value = neMatcher.group(1);
      op2Value = neMatcher.group(2);
      //Check that if op1 and op2 contains quote chars then it should contains a pair number of those chars
      if (areOperandsValid(op1Value, op2Value)) {
        operationType = OperationType.NOT_EQUALS;
      }
    }
    // parse =
    Matcher eMatcher = Pattern.compile("^(.+)=(.+)$").matcher(fromExpr);
    if (operationType == null && eMatcher.find()) {
      op1Value = eMatcher.group(1);
      op2Value = eMatcher.group(2);
      //Check that if op1 and op2 contains quote chars then it should contains a pair number of those chars
      if (areOperandsValid(op1Value, op2Value)) {
        operationType = OperationType.EQUALS;
      }
    }
    // parse OR
    Matcher orMatcher = Pattern.compile("^(.+) or (.+)$").matcher(fromExpr);
    if (operationType == null && orMatcher.find()) {
      op1Value = orMatcher.group(1);
      op2Value = orMatcher.group(2);
      //Check that if op1 and op2 contains quote chars then it should contains a pair number of those chars
      if (areOperandsValid(op1Value, op2Value)) {
        operationType = OperationType.OR;
      }
    }
    // parse AND
    Matcher andMatcher = Pattern.compile("^(.+) and (.+)$").matcher(fromExpr);
    if (operationType == null && andMatcher.find()) {
      op1Value = andMatcher.group(1);
      op2Value = andMatcher.group(2);
      //Check that if op1 and op2 contains quote chars then it should contains a pair number of those chars
      if (areOperandsValid(op1Value, op2Value)) {
        operationType = OperationType.AND;
      }
    }
    // parse - at last as '-' could be part of a number literal value
    Matcher subMatcher = Pattern.compile("^(.+)-(.+)$").matcher(fromExpr);
    if (operationType == null && subMatcher.find()) {
      op1Value = subMatcher.group(1);
      op2Value = subMatcher.group(2);
      //Check that if op1 and op2 contains quote chars then it should contains a pair number of those chars
      if (areOperandsValid(op1Value, op2Value)) {
        operationType = OperationType.SUBTRACT;
      }
    }
    // Create model
    if (operationType != null) {
      AbstractElementWithPosition op1Element = parseXpathExpression(op1Value, fromExpression, dataMappingModel, context, position);
      if (op1Element == null) {
        throw new TransformationException("Invalid first operand in operation " + operationType + ": " + op1Value);
      }
      AbstractElementWithPosition op2Element = parseXpathExpression(op2Value, fromExpression, dataMappingModel, context, position + 2);
      if (op2Element == null) {
        throw new TransformationException("Invalid second operand in operation " + operationType + ": " + op2Value);
      }
      ExpressionOperationModel operationModel = new ExpressionOperationModel();
      operationModel.setType(operationType);
      // TODO layout
      operationModel.setX(250);
      operationModel.setY(50 * (position + 1));
      dataMappingModel.getElements().add(operationModel);

      // Create link between from operand1 to operation
      FlowModel flowModel1 = new FlowModel();
      flowModel1.setSourceElement(op1Element);
      op1Element.getOutgoingFlows().add(flowModel1);
      flowModel1.setTargetElement(operationModel);
      operationModel.getIncomingFlows().add(flowModel1);
      dataMappingModel.getElements().add(flowModel1);
      // Create link between from operand2 to operation
      FlowModel flowModel2 = new FlowModel();
      flowModel2.setSourceElement(op2Element);
      op2Element.getOutgoingFlows().add(flowModel2);
      flowModel2.setTargetElement(operationModel);
      operationModel.getIncomingFlows().add(flowModel2);
      dataMappingModel.getElements().add(flowModel2);
      return operationModel;
    }
    return null;
  }

  public static AbstractElementWithPosition getFromExpressionModel(
          final TExpression fromExpression,
          final DataMappingModel dataMappingModel,
          final Bpmn2DesignerContext context,
          final int position) throws TransformationException {
    String fromExpr = "";
    for (final Object content : fromExpression.getContent()) {
      if (content instanceof String && !"".equals(content)) {
        fromExpr = (String) content;
      }
    }
    return parseXpathExpression(fromExpr, fromExpression, dataMappingModel, context, position);
  }

  private static AbstractElementWithPosition parseXpathExpression(String fromExpr, final TExpression fromExpression, final DataMappingModel dataMappingModel, final Bpmn2DesignerContext context, final int position) throws TransformationException {
    // remove leading and trailing whitespace
    fromExpr = fromExpr.trim();
    //trim parenthesis
    if (fromExpr.startsWith("(") && fromExpr.endsWith(")")) {
      fromExpr = fromExpr.substring(1, fromExpr.length() - 1);
      fromExpr = fromExpr.trim();
    }
    AbstractElementWithPosition result;
    // look for an operation
    result = parseOperation(fromExpr, fromExpression, dataMappingModel, context, position);
    if (result != null) {
      return result;
    }
    // look for getDataObject
    result = parseGetDataObject(fromExpr, fromExpression, dataMappingModel, context, position);
    if (result != null) {
      return result;
    }
    // look for a literal
    result = parseLiteral(fromExpr, fromExpression, dataMappingModel, context, position);
    if (result != null) {
      return result;
    }
    return null;
  }


  public static void bpmnToModel(
          final List<DataInputAssociation> dataInputAssociations,
          final AbstractBPMNElementWithData elementWithData,
          final Bpmn2DesignerContext context) throws TransformationException {
    DataMappingModel dataMappingModel = new DataMappingModel();
    int position = 0;
    for (DataInputAssociation dataInputAssociation : dataInputAssociations) {
      position++;
      bpmnToModel(dataInputAssociation, dataMappingModel, elementWithData, context, position);
    }
    generateLayout(dataMappingModel);
    elementWithData.setDataMappingModel(dataMappingModel);
  }

  public static void getDataInputModels(final AbstractBPMNElementWithData elementWithData,
                                        final List<DataInput> dataInputs,
                                        final Bpmn2DesignerContext context) {
    for (DataInput dataInput : dataInputs) {
      DataInputModel dataInputModel = new DataInputModel();
      dataInputModel.setId(dataInput.getId());
      dataInputModel.setName(dataInput.getName());
      String itemDefinitionId = dataInput.getItemSubjectRef().getLocalPart(); // TODO handle multiple namespaces
      DataTypeModel dataTypeModel = ElementTransformer.getDataTypeModel(context, itemDefinitionId);

      dataInputModel.setType(dataTypeModel);
      dataInputModel.setElementModel(elementWithData);
      elementWithData.getDataInputs().add(dataInputModel);
    }
  }


  public static void generateLayout(DataMappingModel dataMappingModel) {
    List<ExpressionDataInputModel> dataInputModels = new ArrayList<ExpressionDataInputModel>();
    for (AbstractElement element : dataMappingModel.getElements()) {
      if (element instanceof ExpressionDataInputModel) {
        dataInputModels.add((ExpressionDataInputModel) element);
      }
    }
    int yPosition = 20;
    for (ExpressionDataInputModel dataInputModel : dataInputModels) {
      yPosition = generateLayout(dataInputModel, 20, yPosition) + 60;
    }
  }

  private static int generateLayout(AbstractElementWithPosition element, int xPos, int yPos) {
    int maxYPos = yPos;
    int maxXPos = xPos;
    if (!element.getIncomingFlows().isEmpty()) {
      for (FlowModel incomingFlow : element.getIncomingFlows()) {
        AbstractElementWithPosition incomingElement = incomingFlow.getSourceElement();
        maxYPos = generateLayout(incomingElement, xPos, maxYPos) + 60;
        maxXPos = Math.max(maxXPos, incomingElement.getX() + 100);
      }
      maxYPos -= 60;
    }
    element.setX(maxXPos);
    element.setY((maxYPos + yPos) / 2);
    return maxYPos;
  }
}
