/**
 * Bull SAS / OW2 Orchestra
 * Copyright (C) 2010 Bull S.A.S, and individual contributors as indicated
 * by the @authors tag.
 *
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.designer.models.transformers.bpmn2designer;

import java.util.Collection;
import java.util.List;
import javax.wsdl.Definition;
import javax.wsdl.Message;
import javax.wsdl.Part;
import javax.xml.bind.JAXBElement;
import javax.xml.namespace.QName;
import org.apache.ws.commons.schema.XmlSchemaComplexType;
import org.apache.ws.commons.schema.XmlSchemaElement;
import org.apache.ws.commons.schema.XmlSchemaParticle;
import org.apache.ws.commons.schema.XmlSchemaSequence;
import org.apache.ws.commons.schema.XmlSchemaSequenceMember;
import org.apache.ws.commons.schema.XmlSchemaSimpleType;
import org.apache.ws.commons.schema.XmlSchemaType;
import org.ow2.orchestra.designer.bpmn.model.AbstractElement;
import org.ow2.orchestra.designer.bpmn.model.data.DataObjectModel;
import org.ow2.orchestra.designer.bpmn.model.data.type.DataTypeModel;
import org.ow2.orchestra.designer.bpmn.model.data.type.XSDSequenceType;
import org.ow2.orchestra.designer.bpmn.model.data.type.XSDSimpleType;
import org.ow2.orchestra.designer.bpmn.model.data.type.WSDLMessageModel;
import org.ow2.orchestra.designer.bpmn.model.data.type.WSDLMessagePartModel;
import org.ow2.orchestra.designer.bpmn.model.data.type.XSDElementModel;
import org.ow2.orchestra.designer.bpmn.model.data.type.XSDTypeModel;
import org.ow2.orchestra.designer.bpmn.model.process.AbstractBPMNElementWithPosition;
import org.ow2.orchestra.designer.bpmn.model.process.SequenceFlowModel;
import org.ow2.orchestra.designer.models.exceptions.TransformationException;
import org.ow2.orchestra.jaxb.bpmn.TDataObject;
import org.ow2.orchestra.jaxb.bpmn.TEvent;
import org.ow2.orchestra.jaxb.bpmn.TFlowElement;
import org.ow2.orchestra.jaxb.bpmn.TFlowNode;
import org.ow2.orchestra.jaxb.bpmn.TGateway;
import org.ow2.orchestra.jaxb.bpmn.TItemDefinition;
import org.ow2.orchestra.jaxb.bpmn.TRootElement;
import org.ow2.orchestra.jaxb.bpmn.TSequenceFlow;
import org.ow2.orchestra.jaxb.bpmn.TTask;
import org.ow2.orchestra.util.Misc;


/**
 * Utility class used to retrieve the correct transformer for a given object
 *
 * @author Loic Albertin
 */
public final class ElementTransformer {

  /**
   * Utility classes should not have public or default constructors.
   */
  private ElementTransformer() {
  }


  /**
   * Transform an {@link TFlowElement} into a {@link AbstractElement}. this
   * method mainly retrieve the correct transformer and call its bpmnToModel
   * method.
   *
   * @param element the element to transform
   * @param context the transformation context
   * @return Returns true if the transformation succeeded
   * @throws TransformationException Throws {@link TransformationException} if
   *                                 any of sub-transformers throws a {@link TransformationException}
   */
  public static void bpmnToModel(final TFlowElement element, final Bpmn2DesignerContext context) throws TransformationException {
    if (!context.getModelsMap().containsKey(element)) {
      if (element instanceof TGateway) {
        final TGateway tGateway = (TGateway) element;
        GatewayTransformer.bpmnToModel(tGateway, context);
      } else if (element instanceof TSequenceFlow) {
        final TSequenceFlow tSequenceFlow = (TSequenceFlow) element;
        SequenceFlowTransformer.bpmnToModel(tSequenceFlow, context);
      } else if (element instanceof TEvent) {
        final TEvent tEvent = (TEvent) element;
        EventTransformer.bpmnToModel(tEvent, context);
      } else if (element instanceof TTask) {
        final TTask tTask = (TTask) element;
        TaskTransformer.bpmnToModel(tTask, context);
      } else if (element instanceof TDataObject) {
        ElementTransformer.bpmnToModelDataObjects((TDataObject) element, context);
      }
      if (element instanceof TFlowNode) {
        ElementTransformer.bpmnToModelIncomingOutgoingSequenceFlows((TFlowNode) element, context);
      }
    }
    // Designer model represent a sub-part of BPMN model so
    // some BPMN elements doesn't have a transformation into Designer model.
  }

  public static void bpmnToModelIncomingOutgoingSequenceFlows(final TFlowNode bpmnElement, final Bpmn2DesignerContext context) throws TransformationException {
    final AbstractElement contextElement = context.getModelsMap().get(bpmnElement);
    if (contextElement instanceof AbstractBPMNElementWithPosition) {
      final AbstractBPMNElementWithPosition modelElement = (AbstractBPMNElementWithPosition) contextElement;
      final List<QName> incomingSequenceFlowRefs = bpmnElement.getIncomings();
      if (incomingSequenceFlowRefs != null) {
        for (final QName incomingSequenceFlowRef : incomingSequenceFlowRefs) {
          // Get incomingSequenceFlow
          final TFlowElement incomingElement = context.getFlowElementsMap().get(incomingSequenceFlowRef);
          // Generate model for incoming sequenceFlow
          ElementTransformer.bpmnToModel(incomingElement, context);
          // get from bpmn model and add to designer model
          final AbstractElement abstractElement = context.getModelsMap().get(incomingElement);
          if (abstractElement instanceof SequenceFlowModel) {
            modelElement.getIncomingFlows().add((SequenceFlowModel) abstractElement);
          } else {
            throw new TransformationException("BPMN conversion error: incoming element " + incomingSequenceFlowRef + " of " + bpmnElement.getId()
                                              + " does not reference a valid SequenceFlow.");
          }
        }
      }
      final List<QName> outgoingSequenceFlowRefs = bpmnElement.getOutgoings();
      if (outgoingSequenceFlowRefs != null) {
        for (final QName outgoingSequenceFlowRef : outgoingSequenceFlowRefs) {
          // Get SequenceFlow
          final TFlowElement outgoingElement = context.getFlowElementsMap().get(outgoingSequenceFlowRef);
          // Generate model for sequenceFlow
          ElementTransformer.bpmnToModel(outgoingElement, context);
          // get from bpmn model and add to designer model
          final AbstractElement abstractElement = context.getModelsMap().get(outgoingElement);
          if (abstractElement instanceof SequenceFlowModel) {
            modelElement.getOutgoingFlows().add((SequenceFlowModel) abstractElement);
          } else {
            throw new TransformationException("BPMN conversion error: incoming element " + outgoingSequenceFlowRef + " of " + bpmnElement.getId()
                                              + " does not reference a valid SequenceFlow.");
          }
        }
      }
    }
  }

  public static void bpmnToModelDataObjects(final TDataObject dataObject, final Bpmn2DesignerContext context) throws TransformationException {
    if (!context.getModelsMap().containsKey(dataObject)) {
      DataObjectModel dataObjectModel = new DataObjectModel();
      dataObjectModel.setId(dataObject.getId());
      dataObjectModel.setName(dataObject.getName());
      if (dataObject.getItemSubjectRef() == null) {
        throw new TransformationException("No datatype defined for dataObject " + dataObject.getName());
      }
      String itemDefinitionId = dataObject.getItemSubjectRef().getLocalPart(); // TODO handle multiple namespaces
      dataObjectModel.setType(ElementTransformer.getDataTypeModel(context, itemDefinitionId));

      dataObjectModel.setProcessModel(context.getProcessModel());
      context.getProcessModel().getDataObjects().put(dataObjectModel.getName(), dataObjectModel);
      context.getModelsMap().put(dataObject, dataObjectModel);
    }
  }

  public static DataTypeModel getDataTypeModel(final Bpmn2DesignerContext context, final String itemDefinitionId) {
    // get data type
    for (final JAXBElement<? extends TRootElement> rootElement : context.getDefinitions().getRootElements()) {
      if (rootElement.getValue() instanceof TItemDefinition
          && rootElement.getValue().getId().equals(itemDefinitionId)) {
        QName structureQName = ((TItemDefinition) rootElement.getValue()).getStructureRef();
        return getDataTypeModel(context, structureQName);
      }
    }
    return null;
  }

  private static DataTypeModel getDataTypeModel(final Bpmn2DesignerContext context, final QName structureQName) {
    // check if structure is a WSDL message
    for (Definition wsdlDefinition : context.getWsdlDefinitions().values()) {
      Message wsdlMessage = wsdlDefinition.getMessage(structureQName);
      if (wsdlMessage != null) {
        if (!context.getDataTypeModelMap().containsKey(structureQName)) {
          WSDLMessageModel wsdlMessageModel = new WSDLMessageModel();
          wsdlMessageModel.setQName(structureQName);
          wsdlMessageModel.setId(Misc.getUniqueId("_"));
          getWSDLMessagePartModel(context, wsdlMessageModel, wsdlMessage);
          context.getDataTypeModelMap().put(structureQName, wsdlMessageModel);
        }
        return context.getDataTypeModelMap().get(structureQName);
      }
    }
    // check if structure is a XSD element
    XmlSchemaElement xmlSchemaElement = context.getXmlSchemaCollection().getElementByQName(structureQName);
    if (xmlSchemaElement != null) {
      if (xmlSchemaElement.isTopLevel()) {
        if (!context.getDataTypeModelMap().containsKey(structureQName)) {
          XSDElementModel xsdElementModel = getXSDElementModel(xmlSchemaElement);
          context.getDataTypeModelMap().put(structureQName, xsdElementModel);
        }
        return context.getDataTypeModelMap().get(structureQName);
      } else {
        XSDElementModel xsdElementModel = getXSDElementModel(xmlSchemaElement);
        return xsdElementModel;
      }
    }
    // structure is a XSD type
    XmlSchemaType xmlSchemaType = context.getXmlSchemaCollection().getTypeByQName(structureQName);
    if (xmlSchemaType != null) {
      if (xmlSchemaType.isTopLevel()) {
        if (!context.getDataTypeModelMap().containsKey(structureQName)) {
          context.getDataTypeModelMap().put(structureQName, getXSDTypeModel(xmlSchemaType));
        }
        return context.getDataTypeModelMap().get(structureQName);
      } else {
        return getXSDTypeModel(xmlSchemaType);
      }
    }
    // TODO Nothing found. Throw exception ?
    XSDTypeModel typeModel = new XSDSimpleType();
    typeModel.setQName(structureQName);
    typeModel.setId(Misc.getUniqueId("_"));

    return typeModel;
  }

  private static XSDElementModel getXSDElementModel(final XmlSchemaElement xmlSchemaElement) {
    XSDElementModel xsdElementModel = new XSDElementModel();
    xsdElementModel.setQName(xmlSchemaElement.getQName());
    xsdElementModel.setId(Misc.getUniqueId("_"));
    if (xmlSchemaElement.getSchemaTypeName() != null) {
      XSDTypeModel typeModel = new XSDSimpleType();
      typeModel.setQName(xmlSchemaElement.getSchemaTypeName());
      typeModel.setId(Misc.getUniqueId("_"));
      xsdElementModel.setType(typeModel);
    } else {
      xsdElementModel.setType(getXSDTypeModel(xmlSchemaElement.getSchemaType()));
    }
    return xsdElementModel;
  }

  private static XSDTypeModel getXSDTypeModel(final XmlSchemaType schemaType) {
    if (schemaType instanceof XmlSchemaSimpleType) {
      XSDTypeModel typeModel = new XSDSimpleType();
      typeModel.setQName(schemaType.getQName());
      typeModel.setId(Misc.getUniqueId("_"));
    } else {
      XmlSchemaComplexType schemaComplexType = (XmlSchemaComplexType) schemaType;
      XmlSchemaParticle schemaParticle = schemaComplexType.getParticle();
      if (schemaParticle instanceof XmlSchemaSequence) {
        XmlSchemaSequence schemaSequence = (XmlSchemaSequence) schemaParticle;
        XSDSequenceType xsdSequenceType = new XSDSequenceType();
        xsdSequenceType.setId(Misc.getUniqueId("_"));
        if (schemaComplexType.getQName() != null) {
          xsdSequenceType.setQName(schemaComplexType.getQName());
        } else {
          xsdSequenceType.setQName(new QName(schemaComplexType.getParent().getTargetNamespace(), xsdSequenceType.getId()));
        }
        for (XmlSchemaSequenceMember seqMember : schemaSequence.getItems()) {
          if (seqMember instanceof XmlSchemaElement) {
            xsdSequenceType.getElements().add(getXSDElementModel((XmlSchemaElement) seqMember));
          }
        }
        return xsdSequenceType;
      }
    }
    // TODO Nothing found. Throw exception ?
    XSDTypeModel typeModel = new XSDSimpleType();
    typeModel.setQName(schemaType.getQName());
    typeModel.setId(Misc.getUniqueId("_"));
    return typeModel;
  }

  public static void getWSDLMessagePartModel(
          final Bpmn2DesignerContext context,
          final WSDLMessageModel wsdlMessageModel,
          final Message wsdlMessage) {
    // parse WSDL message from wsdls
    for (Part part : (Collection<Part>) wsdlMessage.getParts().values()) {
      final WSDLMessagePartModel wsdlMessagePartModel = new WSDLMessagePartModel();
      wsdlMessagePartModel.setPartName(part.getName());
      if (part.getTypeName() != null) {
        XSDTypeModel xsdType = (XSDTypeModel) getDataTypeModel(context, part.getTypeName());
        wsdlMessagePartModel.setXsdType(xsdType);
      } else {
        XSDElementModel xsdElement = (XSDElementModel) getDataTypeModel(context, part.getElementName());
        wsdlMessagePartModel.setXsdElement(xsdElement);
      }
      wsdlMessageModel.getParts().add(wsdlMessagePartModel);
    }
  }
}
