/**
 * Bull SAS / OW2 Orchestra
 * Copyright (C) 2010 Bull S.A.S, and individual contributors as indicated
 * by the @authors tag.
 *
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.designer.models.transformers.bpmn2designer;

import java.util.List;
import javax.xml.bind.JAXBElement;
import javax.xml.namespace.QName;
import org.ow2.orchestra.designer.bpmn.model.data.DataObjectModel;
import org.ow2.orchestra.designer.bpmn.model.operations.InterfaceModel;
import org.ow2.orchestra.designer.bpmn.model.operations.OperationModel;
import org.ow2.orchestra.designer.bpmn.model.process.EventModel;
import org.ow2.orchestra.designer.bpmn.model.process.EventType;
import org.ow2.orchestra.designer.models.exceptions.TransformationException;
import org.ow2.orchestra.jaxb.bpmn.DataOutputAssociation;
import org.ow2.orchestra.jaxb.bpmn.MessageFlow;
import org.ow2.orchestra.jaxb.bpmn.Operation;
import org.ow2.orchestra.jaxb.bpmn.TCatchEvent;
import org.ow2.orchestra.jaxb.bpmn.TDataObject;
import org.ow2.orchestra.jaxb.bpmn.TEvent;
import org.ow2.orchestra.jaxb.bpmn.TEventDefinition;
import org.ow2.orchestra.jaxb.bpmn.TInterface;
import org.ow2.orchestra.jaxb.bpmn.TMessageEventDefinition;
import org.ow2.orchestra.jaxb.bpmn.TRootElement;
import org.ow2.orchestra.jaxb.bpmn.TThrowEvent;
import org.ow2.orchestra.jaxb.bpmn.TTimerEventDefinition;

/**
 * Transform {@link TEvent}s into {@link EventModel}s and vice versa
 *
 * @author Guillaume Porcher
 */
public final class EventTransformer {

  /**
   * Utility classes should not have public or default constructors.
   */
  private EventTransformer() {

  }

  /**
   * @param tEvent
   * @param context
   * @return
   * @throws TransformationException
   */
  public static void bpmnToModel(final TEvent tEvent, final Bpmn2DesignerContext context) throws TransformationException {
    final EventModel eventModel = new EventModel();
    eventModel.setName(tEvent.getName());
    eventModel.setProcessModel(context.getProcessModel());
    context.getProcessModel().getElements().add(eventModel);
    context.getModelsMap().put(tEvent, eventModel);

    if (tEvent.getId() == null || "".equals(tEvent.getId())) {
      // If there is no id, we will not be able to find out a diagram element
      // for this gateway.
      // We will not be able to reference this gateway from the designer, so this is an error
      throw new TransformationException("BPMN conversion error: An event has no id.");
    }
    eventModel.setId(tEvent.getId());

    // retrieve event definition
    final List<JAXBElement<? extends TEventDefinition>> eventDefinitions;
    if (tEvent instanceof TThrowEvent) {
      eventDefinitions = ((TThrowEvent) tEvent).getEventDefinitions();
    } else if (tEvent instanceof TCatchEvent) {
      eventDefinitions = ((TCatchEvent) tEvent).getEventDefinitions();
    } else {
      eventDefinitions = null;
    }
    if (eventDefinitions != null && eventDefinitions.size() > 1) {
      throw new TransformationException("BPMN conversion error: event " + tEvent.getId() + " has multiple eventDefinitions. This is not supported.");
    }

    if (eventDefinitions != null && eventDefinitions.size() == 1) {
      final TEventDefinition eventDefinition = eventDefinitions.get(0).getValue();
      if (eventDefinition instanceof TTimerEventDefinition) {
        // Tmier event
        final TTimerEventDefinition timerDefinition = (TTimerEventDefinition) eventDefinition;
        eventModel.setEventType(EventType.TIMER);
        if (timerDefinition.getTimeDate() != null && !timerDefinition.getTimeDate().getContent().isEmpty()) {
          eventModel.setTimerDate(timerDefinition.getTimeDate().getContent().get(0).toString());
        } else if (timerDefinition.getTimeDuration() != null && !timerDefinition.getTimeDuration().getContent().isEmpty()) {
          eventModel.setTimerDuration(timerDefinition.getTimeDuration().getContent().get(0).toString());
        }
      } else if (eventDefinition instanceof TMessageEventDefinition) {
        // Message Event
        if (tEvent instanceof TThrowEvent) {
          eventModel.setEventType(EventType.SEND_MESSAGE);
        } else if (tEvent instanceof TCatchEvent) {
          eventModel.setEventType(EventType.RECEIVE_MESSAGE);
        }
        // Get operation name and interface name
        final TMessageEventDefinition messageEventDefinition = (TMessageEventDefinition) eventDefinition;
        final QName operationQName = messageEventDefinition.getOperationRef();
        if (operationQName == null || operationQName.getNamespaceURI() == null || "".equals(operationQName.getNamespaceURI())) {
          throw new TransformationException("BPMN conversion error: event " + tEvent.getId() + " doesn't reference any " +
                                            "operation.");
        }
        if (!operationQName.getNamespaceURI().equals(context.getDefinitions().getTargetNamespace())) {
          throw new TransformationException("BPMN conversion error: event " + tEvent.getId() + " references an unknown operation.");
        }
        TInterface tInterface = null;
        Operation operation = null;
        for (final JAXBElement<? extends TRootElement> rootElement : context.getDefinitions().getRootElements()) {
          if (rootElement.getValue() instanceof TInterface) {
            for (final Operation interfaceOperation : ((TInterface) rootElement.getValue()).getOperations()) {
              if (interfaceOperation.getId().equals(operationQName.getLocalPart())) {
                tInterface = (TInterface) rootElement.getValue();
                operation = interfaceOperation;
              }
            }
          }
        }
        if (operation == null) {
          throw new TransformationException("BPMN conversion error: event " + tEvent.getId() + " references an unknown operation.");
        }
        final InterfaceModel interfaceModel = context.getProcessModel().getInterfaces().get(tInterface.getName());
        if (interfaceModel == null) {
          throw new TransformationException("BPMN conversion error: event " + tEvent.getId() + " references an unknown operation.");
        }
        final OperationModel operationModel = interfaceModel.getOperations().get(operation.getName());
        if (operationModel == null) {
          throw new TransformationException("BPMN conversion error: event " + tEvent.getId() + " references an unknown operation.");
        }
        eventModel.setOperation(operationModel);
        if (tEvent instanceof TThrowEvent) {
          operationModel.getSendElements().add(eventModel);
        } else if (tEvent instanceof TCatchEvent) {
          operationModel.getReceiveElements().add(eventModel);
        }
        // get correlations
        MessageFlow messageFlow = null;
        if (tEvent instanceof TThrowEvent) {
          messageFlow = context.getOutgoingMessageFlow(tEvent);
        } else if (tEvent instanceof TCatchEvent) {
          messageFlow = context.getIncomingMessageFlow(tEvent);
        }
        eventModel.getCorrelationMappingModels().addAll(CorrelationTransformer.bpmnToModelCorrelationMapping(messageFlow, context));

        if (tEvent instanceof TThrowEvent) {
          // get data inputs
          DataMapperTransformer.getDataInputModels(eventModel, ((TThrowEvent) tEvent).getDataInputs(), context);
          // Get expression
          DataMapperTransformer.bpmnToModel(((TThrowEvent) tEvent).getDataInputAssociations(), eventModel, context);
        } else if (tEvent instanceof TCatchEvent) {
          // Get variables
          for (final DataOutputAssociation dataOutputAssociation : ((TCatchEvent) tEvent).getDataOutputAssociations()) {
            if (dataOutputAssociation.getTargetRef() instanceof TDataObject) {
              ElementTransformer.bpmnToModel((TDataObject) dataOutputAssociation.getTargetRef(), context);
              DataObjectModel variable = (DataObjectModel) context.getModelsMap().get(dataOutputAssociation.getTargetRef());
              eventModel.getOutputVariables().add(variable);
              variable.getUsedBy().add(eventModel);
            }
          }
        }
      }
    }

    // Now retrieve positions
    ShapeTransformer.bpmnToModel(tEvent.getId(), eventModel, context);
  }

}
