/**
 * Bull SAS / OW2 Orchestra
 * Copyright (C) 2010 Bull S.A.S, and individual contributors as indicated
 * by the @authors tag.
 *
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.designer.models.transformers.bpmn2designer;

import javax.xml.XMLConstants;
import javax.xml.namespace.QName;
import org.ow2.orchestra.designer.bpmn.model.AbstractElementWithPosition;
import org.ow2.orchestra.designer.bpmn.model.FlowModel;
import org.ow2.orchestra.designer.bpmn.model.data.DataInputModel;
import org.ow2.orchestra.designer.bpmn.model.data.DataMappingModel;
import org.ow2.orchestra.designer.bpmn.model.data.ExpressionDataInputModel;
import org.ow2.orchestra.designer.bpmn.model.data.type.XSDSimpleType;
import org.ow2.orchestra.designer.bpmn.model.process.AbstractBPMNElementWithPosition;
import org.ow2.orchestra.designer.bpmn.model.process.SequenceFlowModel;
import org.ow2.orchestra.designer.models.exceptions.TransformationException;
import org.ow2.orchestra.jaxb.bpmn.TExclusiveGateway;
import org.ow2.orchestra.jaxb.bpmn.TFlowElement;
import org.ow2.orchestra.jaxb.bpmn.TSequenceFlow;
import org.ow2.orchestra.util.Misc;

/**
 * Transforms {@link SequenceFlowModel}s into {@link TSequenceFlow}s and vice
 * versa
 *
 * @author Loic Albertin
 */
public final class SequenceFlowTransformer {

  /**
   * Utility classes should not have public or default constructors.
   */
  private SequenceFlowTransformer() {

  }

  /**
   * @param tSequenceFlow
   * @param context
   * @return
   * @throws TransformationException
   */
  public static void bpmnToModel(final TSequenceFlow tSequenceFlow, final Bpmn2DesignerContext context)
          throws TransformationException {
    final SequenceFlowModel sequenceFlowModel = new SequenceFlowModel();
    sequenceFlowModel.setName(tSequenceFlow.getName());
    sequenceFlowModel.setId(tSequenceFlow.getId());

    // Checks source and target
    if (!(tSequenceFlow.getSourceRef() instanceof TFlowElement)) {
      throw new TransformationException("BPMN conversion error: sourceRef attribute of sequenceFlow " + tSequenceFlow.getId()
                                        + " is not a FlowElement.");
    }

    if (!(tSequenceFlow.getTargetRef() instanceof TFlowElement)) {
      throw new TransformationException("BPMN conversion error: targetRef attribute of sequenceFlow " + tSequenceFlow.getId()
                                        + " is not a FlowElement.");
    }
    // Add to list before calling ElementTransformer.bpmnToModel (can cause recursion and duplicate elements)
    context.getModelsMap().put(tSequenceFlow, sequenceFlowModel);

    final TFlowElement sourceElement = (TFlowElement) tSequenceFlow.getSourceRef();
    ElementTransformer.bpmnToModel(sourceElement, context);

    final TFlowElement targetElement = (TFlowElement) tSequenceFlow.getTargetRef();
    ElementTransformer.bpmnToModel(targetElement, context);

    if (!(context.getModelsMap().get(tSequenceFlow.getSourceRef()) instanceof AbstractBPMNElementWithPosition)) {
      throw new TransformationException("BPMN conversion error: sourceRef attribute of sequenceFlow " + tSequenceFlow.getId()
                                        + " is not a valid FlowElement.");
    }

    if (!(context.getModelsMap().get(tSequenceFlow.getTargetRef()) instanceof AbstractBPMNElementWithPosition)) {
      throw new TransformationException("BPMN conversion error: targetRef attribute of sequenceFlow " + tSequenceFlow.getId()
                                        + " is not a valid FlowElement.");
    }
    sequenceFlowModel.setSourceElement((AbstractBPMNElementWithPosition) context.getModelsMap().get(tSequenceFlow.getSourceRef()));
    sequenceFlowModel.setTargetElement((AbstractBPMNElementWithPosition) context.getModelsMap().get(tSequenceFlow.getTargetRef()));

    context.getProcessModel().getElements().add(sequenceFlowModel);

    // Parse condition expression
    if (tSequenceFlow.getConditionExpression() != null
        && !tSequenceFlow.getConditionExpression().getContent().isEmpty()) {
      sequenceFlowModel.setHasCondition(true);

      // Create DataMapping from XpathExpression
      DataMappingModel dataMappingModel = new DataMappingModel();
      // Parse from expr
      AbstractElementWithPosition fromExpressionModel =
              DataMapperTransformer.getFromExpressionModel(tSequenceFlow.getConditionExpression(), dataMappingModel, context, 1);
      if (fromExpressionModel == null) {
        throw new TransformationException("Unable to generate expression.");
      }
      // Create link between from expr and to expr
      FlowModel flowModel = new FlowModel();
      flowModel.setSourceElement(fromExpressionModel);
      fromExpressionModel.getOutgoingFlows().add(flowModel);

      ExpressionDataInputModel expressionDataInputModel = new ExpressionDataInputModel();

      DataInputModel dataInputModel = new DataInputModel();
      dataInputModel.setName("condition");
      XSDSimpleType booleanType = new XSDSimpleType();
      booleanType.setId(Misc.getUniqueId("_"));
      booleanType.setQName(new QName(XMLConstants.W3C_XML_SCHEMA_NS_URI, "boolean"));
      dataInputModel.setType(booleanType);
      sequenceFlowModel.getDataInputs().add(dataInputModel);

      // set data input
      expressionDataInputModel.setDataInputModel(dataInputModel);
      dataMappingModel.getElements().add(expressionDataInputModel);
      flowModel.setTargetElement(expressionDataInputModel);
      expressionDataInputModel.getIncomingFlows().add(flowModel);
      dataMappingModel.getElements().add(flowModel);

      DataMapperTransformer.generateLayout(dataMappingModel);
      sequenceFlowModel.setDataMappingModel(dataMappingModel);
    }

    // Parse isDefault
    if (sourceElement instanceof TExclusiveGateway) {
      final TExclusiveGateway sourceExclusiveGateway = (TExclusiveGateway) sourceElement;
      if (tSequenceFlow.equals(sourceExclusiveGateway.getDefault())) {
        sequenceFlowModel.setIsDefault(true);
      }
    }
    // Currently we don't care about the diagram part of the sequence flow
  }

}
