/**
 * Bull SAS / OW2 Orchestra
 * Copyright (C) 2010 Bull S.A.S, and individual contributors as indicated
 * by the @authors tag.
 *
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.designer.models.transformers.designer2bpmn;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import javax.xml.XMLConstants;
import javax.xml.namespace.QName;
import org.ow2.orchestra.designer.bpmn.model.AbstractElement;
import org.ow2.orchestra.designer.bpmn.model.AbstractElementWithPosition;
import org.ow2.orchestra.designer.bpmn.model.FlowModel;
import org.ow2.orchestra.designer.bpmn.model.data.DataMappingModel;
import org.ow2.orchestra.designer.bpmn.model.data.DataObjectModel;
import org.ow2.orchestra.designer.bpmn.model.data.ExpressionDataInputModel;
import org.ow2.orchestra.designer.bpmn.model.data.ExpressionDataObjectModel;
import org.ow2.orchestra.designer.bpmn.model.data.ExpressionOperationModel;
import org.ow2.orchestra.designer.bpmn.model.data.LiteralModel;
import org.ow2.orchestra.designer.models.exceptions.TransformationException;
import org.ow2.orchestra.jaxb.bpmn.Assignment;
import org.ow2.orchestra.jaxb.bpmn.DataInput;
import org.ow2.orchestra.jaxb.bpmn.DataInputAssociation;
import org.ow2.orchestra.jaxb.bpmn.TDataObject;
import org.ow2.orchestra.jaxb.bpmn.TExpression;
import org.ow2.orchestra.jaxb.bpmn.TFormalExpression;
import org.ow2.orchestra.util.Misc;

/**
 * Transform {@link org.ow2.orchestra.designer.bpmn.model.data.DataMappingModel}s into {@link org.ow2.orchestra.jaxb.bpmn.DataInputAssociation}
 *
 * @author Guillaume Porcher
 */
public final class DataMapperTransformer {

  /**
   * Utility classes should not have public or default constructors.
   */
  private DataMapperTransformer() {
  }

  /**
   * @param dataMappingModel
   * @return
   */
  public static List<DataInputAssociation> modelToBPMN(final DataMappingModel dataMappingModel,
                                                       final List<DataInput> dataInputs,
                                                       final Map<String, DataObjectModel> dataObjectModels,
                                                       final Designer2BpmnContext context) throws TransformationException {
    List<DataInputAssociation> dataInputAssociations = new ArrayList<DataInputAssociation>();
    if (dataMappingModel == null) {
      return dataInputAssociations;
    }
    // Get data inputs form the data mapping model
    List<ExpressionDataInputModel> dataInputModels = new ArrayList<ExpressionDataInputModel>();
    for (AbstractElement element : dataMappingModel.getElements()) {
      if (element instanceof ExpressionDataInputModel) {
        ExpressionDataInputModel expressionDataInputModel = (ExpressionDataInputModel) element;
        dataInputModels.add(expressionDataInputModel);
      }
    }
    // Generate association for each data input
    for (ExpressionDataInputModel dataInputModel : dataInputModels) {
      DataInputAssociation dataInputAssociation = new DataInputAssociation();
      Assignment assignment = new Assignment();
      TFormalExpression toExpression = new TFormalExpression();
      String dataInputName = dataInputModel.getDataInputModel().getName();
      if (dataInputModel.getChildNodeSelector() == null) {
        toExpression.getContent().add("getDataInput(\"" + dataInputName + "\")");
      } else {
        QName childNodeQName = QName.valueOf(dataInputModel.getChildNodeSelector());
        String prefix = Misc.getHumanReadableId("ns");
        toExpression.getOtherAttributes().put(new QName(XMLConstants.XMLNS_ATTRIBUTE_NS_URI, prefix), childNodeQName.getNamespaceURI());
        toExpression.getContent().add("getDataInput(\"" + dataInputName + "\")/" + prefix + ":" + childNodeQName.getLocalPart());
      }
      assignment.setTo(toExpression);

      DataInput dataInput = null;
      for (DataInput di : dataInputs) {
        if (di.getName().equals(dataInputName)) {
          dataInput = di;
        }
      }
      if (dataInput == null) {
        throw new TransformationException("No dataInput " + dataInputName + " declared in element");
      }
      dataInputAssociation.setTargetRef(dataInput);

      TFormalExpression fromExpression = new TFormalExpression();
      fromExpression.getContent().add(getXPathExpression(dataInputModel, dataObjectModels, dataInputAssociation, fromExpression, context));
      assignment.setFrom(fromExpression);
      dataInputAssociation.getAssignments().add(assignment);
      dataInputAssociations.add(dataInputAssociation);
    }
    return dataInputAssociations;
  }

  private static String getXPathExpression(
          final AbstractElementWithPosition expressionElementModel,
          final Map<String, DataObjectModel> dataObjectModels,
          final DataInputAssociation dataInputAssociation,
          final TExpression targetExpression,
          final Designer2BpmnContext context) throws TransformationException {
    if (expressionElementModel instanceof ExpressionDataInputModel) {
      // return (unique) source XPath expression
      if (expressionElementModel.getIncomingFlows().size() != 1) {
        throw new TransformationException("Invalid incoming flow number in data mapping");
      }
      FlowModel incomingFlow = expressionElementModel.getIncomingFlows().get(0);
      return getXPathExpression(
              incomingFlow.getSourceElement(),
              dataObjectModels,
              dataInputAssociation,
              targetExpression,
              context);
    } else if (expressionElementModel instanceof ExpressionDataObjectModel) {
      ExpressionDataObjectModel elementModel = (ExpressionDataObjectModel) expressionElementModel;
      final String dataObjectName = elementModel.getDataObjectModel().getName();

      DataObjectModel dataObjectModel = dataObjectModels.get(dataObjectName);
      if (dataObjectModel == null) {
        throw new TransformationException("No dataObject " + dataObjectName + " declared in the process");
      }
      ElementTransformer.modelToBpmnDataObject(context, dataObjectModel);
      TDataObject dataObject = (TDataObject) context.getModelsMap().get(dataObjectModel);
      dataInputAssociation.getSourceReves().add(context.getBpmnObjectFactory().createTDataAssociationSourceRef(dataObject));

      if (elementModel.getChildNodeSelector() == null) {
        return "getDataObject(\"" + dataObjectName + "\")";
      }
      QName childNodeQName = QName.valueOf(elementModel.getChildNodeSelector());
      String prefix = Misc.getHumanReadableId("ns");
      targetExpression.getOtherAttributes().put(new QName(XMLConstants.XMLNS_ATTRIBUTE_NS_URI, prefix), childNodeQName.getNamespaceURI());
      return "getDataObject(\"" + dataObjectName + "\")/" + prefix + ":" + childNodeQName.getLocalPart();
    } else if (expressionElementModel instanceof LiteralModel) {
      LiteralModel elementModel = (LiteralModel) expressionElementModel;
      final String value = elementModel.getValue();
      if (elementModel.getType() != null && (
              (new QName(XMLConstants.W3C_XML_SCHEMA_NS_URI, "string")).equals(elementModel.getType().getQName()) ||
              (new QName(XMLConstants.W3C_XML_SCHEMA_NS_URI, "dateTime")).equals(elementModel.getType().getQName()))) {
        return "\"" + value + "\"";
      } else if (elementModel.getType() != null &&
                 (new QName(XMLConstants.W3C_XML_SCHEMA_NS_URI, "boolean")).equals(elementModel.getType().getQName())) {
        if ("true".equalsIgnoreCase(value)) {
          return "true()";
        } else {
          return "false()";
        }
      } else {
        return value;
      }
    } else if (expressionElementModel instanceof ExpressionOperationModel) {
      ExpressionOperationModel operationModel = (ExpressionOperationModel) expressionElementModel;
      if (operationModel.getIncomingFlows().size() != 2) {
        throw new TransformationException("Error in data mapping: operation " + operationModel.getType() + " needs 2 operands but has " + operationModel.getIncomingFlows().size());
      }
      String operand1 = getXPathExpression(operationModel.getIncomingFlows().get(0).getSourceElement(), dataObjectModels, dataInputAssociation, targetExpression, context);
      String operand2 = getXPathExpression(operationModel.getIncomingFlows().get(1).getSourceElement(), dataObjectModels, dataInputAssociation, targetExpression, context);
      switch (operationModel.getType()) {
        case CONCATENATE:
          return "concat(" + operand1 + "," + operand2 + ")";
        case ADD:
          return "(" + operand1 + " + " + operand2 + ")";
        case SUBTRACT:
          return "(" + operand1 + " - " + operand2 + ")";
        case DIVIDE:
          return "(" + operand1 + " div " + operand2 + ")";
        case MULTIPLY:
          return "(" + operand1 + " * " + operand2 + ")";
        case IS_GREATER_THAN:
          return "(" + operand1 + " > " + operand2 + ")";
        case IS_GREATER_OR_EQUALS:
          return "(" + operand1 + " >= " + operand2 + ")";
        case IS_LESS_THAN:
          return "(" + operand1 + " < " + operand2 + ")";
        case IS_LESS_OR_EQUALS:
          return "(" + operand1 + " <= " + operand2 + ")";
        case NOT_EQUALS:
          return "(" + operand1 + " != " + operand2 + ")";
        case EQUALS:
          return "(" + operand1 + " = " + operand2 + ")";
        case AND:
          return "(" + operand1 + " and " + operand2 + ")";
        case OR:
          return "(" + operand1 + " or " + operand2 + ")";
      }
    }
    throw new TransformationException("Invalid type in data mapping: " + expressionElementModel.getClass().getName());
  }

}
