/**
 * Bull SAS / OW2 Orchestra
 * Copyright (C) 2010 Bull S.A.S, and individual contributors as indicated
 * by the @authors tag.
 *
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.designer.models.transformers.designer2bpmn;

import java.util.Collection;
import javax.xml.bind.JAXBElement;
import javax.xml.namespace.QName;
import org.ow2.orchestra.designer.bpmn.model.AbstractElement;
import org.ow2.orchestra.designer.bpmn.model.data.type.WSDLMessageModel;
import org.ow2.orchestra.designer.bpmn.model.operations.InterfaceModel;
import org.ow2.orchestra.designer.bpmn.model.operations.OperationModel;
import org.ow2.orchestra.designer.bpmn.model.process.PoolModel;
import org.ow2.orchestra.designer.bpmn.model.process.ProcessModel;
import org.ow2.orchestra.designer.models.exceptions.TransformationException;
import org.ow2.orchestra.jaxb.bpmn.Definitions;
import org.ow2.orchestra.jaxb.bpmn.Import;
import org.ow2.orchestra.jaxb.bpmn.Operation;
import org.ow2.orchestra.jaxb.bpmn.Participant;
import org.ow2.orchestra.jaxb.bpmn.TCollaboration;
import org.ow2.orchestra.jaxb.bpmn.TInterface;
import org.ow2.orchestra.jaxb.bpmn.TItemDefinition;
import org.ow2.orchestra.jaxb.bpmn.TMessage;
import org.ow2.orchestra.jaxb.bpmn.TProcess;
import org.ow2.orchestra.jaxb.bpmn.TRootElement;
import org.ow2.orchestra.jaxb.bpmn.bpmndi.BPMNDiagram;
import org.ow2.orchestra.jaxb.bpmn.bpmndi.BPMNPlane;

/**
 * @author Loic Albertin
 */
public final class Designer2BpmnTransformer {

  /**
   * Transforms a designer {@link ProcessModel} into a BPMN {@link Definitions}
   *
   * @param processModel The {@link ProcessModel} to modelToBpmn
   * @return Returns a corresponding {@link Definitions}
   * @throws TransformationException
   */
  public static Definitions modelToBpmn(final ProcessModel processModel) throws TransformationException {

    final Designer2BpmnContext context = new Designer2BpmnContext();

    if (processModel.getId() == null || processModel.getId().equals("")) {
      throw new TransformationException("BPMN conversion error: Process has no id.");
    }

    final Definitions definitions = new Definitions();
    if (processModel.getWsdlNamespace() == null || processModel.getWsdlNamespace().equals("")) {
      throw new TransformationException("BPMN conversion error: Process has no namespace.");
    }
    definitions.setTargetNamespace(processModel.getWsdlNamespace());
    definitions.setId("definition" + processModel.getId());
    definitions.setName(processModel.getName());
    context.setDefinitions(definitions);

    // Initiate Process part
    context.setTProcess(new TProcess());
    context.getTProcess().setName(processModel.getName());
    context.getTProcess().setId(processModel.getId());
    context.getTProcess().setIsExecutable(true);

    definitions.getRootElements().add(context.getBpmnObjectFactory().createProcess(context.getTProcess()));

    // Initiate diagram part
    final BPMNDiagram diagram = new BPMNDiagram();
    diagram.setId("diagram" + processModel.getId());
    diagram.setName(processModel.getName() + "_Diagram");
    final BPMNPlane plane = new BPMNPlane();
    plane.setId("plane" + processModel.getId());
    plane.setBpmnElement(new QName(definitions.getTargetNamespace(), context.getTProcess().getId()));
    diagram.setBPMNPlane(plane);
    context.setBpmnPlane(plane);
    definitions.getBPMNDiagrams().add(diagram);

    Designer2BpmnTransformer.generateProcessInterfaces(processModel, context);
    PoolTransformer.modelToBpmnClientPool(context);

    // Generate other participants
    for (final AbstractElement element : processModel.getElements()) {
      if (element instanceof PoolModel) {
        PoolTransformer.modelToBpmn((PoolModel) element, context);
      }
    }
    // Generate activities
    for (final AbstractElement element : processModel.getElements()) {
      ElementTransformer.modelToBpmn(element, context);
    }

    return definitions;
  }

  /**
   * Generate BPMN classes for the process interfaces.
   * Generates collaboration, process participant, interfaces, operations, messages and messages item definitions.
   *
   * @param processModel
   * @param context
   */
  protected static void generateProcessInterfaces(final ProcessModel processModel, final Designer2BpmnContext context)
          throws TransformationException {
    // Generate collaboration
    if (context.getCollaboration() == null) {
      final TCollaboration collaboration = new TCollaboration();
      context.setCollaboration(collaboration);
      context.getDefinitions().getRootElements().add(context.getBpmnObjectFactory().createCollaboration(collaboration));
    }
    final TCollaboration collaboration = context.getCollaboration();
    // Generate process participant
    Participant processParticipant = null;
    for (final Participant participant : collaboration.getParticipants()) {
      if (new QName(context.getDefinitions().getTargetNamespace(), context.getTProcess().getId()).equals(participant.getProcessRef())) {
        processParticipant = participant;
      }
    }
    if (processParticipant == null) {
      processParticipant = new Participant();
      processParticipant.setProcessRef(new QName(context.getDefinitions().getTargetNamespace(), context.getTProcess().getId()));
      processParticipant.setName(context.getTProcess().getName() + "Participant");
      processParticipant.setId("participant" + context.getTProcess().getId());
      collaboration.getParticipants().add(processParticipant);

      // Now deals with positions
      ShapeTransformer.modelToBpmn(processModel, processParticipant.getId(), context);
    }
    Collection<InterfaceModel> interfaces = processModel.getInterfaces().values();
    // Generate interfaces
    generateInterfaces(context, processParticipant, interfaces);

    // Generate import for process WSDl (generated outside this method)
    Import bpmnImport = new Import();
    bpmnImport.setImportType("http://schemas.xmlsoap.org/wsdl/");
    bpmnImport.setLocation(processModel.getName() + ".wsdl");
    bpmnImport.setNamespace(processModel.getWsdlNamespace());
    context.getDefinitions().getImports().add(bpmnImport);
  }

  public static void generateInterfaces(Designer2BpmnContext context, Participant participant, Collection<InterfaceModel> interfaces) {
    // Generate interfaces
    for (final InterfaceModel interfaceModel : interfaces) {
      final String interfaceName = interfaceModel.getName();
      // Find or create interface
      TInterface tInterface = null;
      for (final JAXBElement<? extends TRootElement> rootElement : context.getDefinitions().getRootElements()) {
        if (rootElement.getValue() instanceof TInterface && ((TInterface) rootElement.getValue()).getName().equals(interfaceName)) {
          tInterface = (TInterface) rootElement.getValue();
        }
      }
      if (tInterface == null) {
        tInterface = new TInterface();
        tInterface.setId(interfaceModel.getId());
        tInterface.setName(interfaceName);
        if (interfaceModel.getPortTypeQName() != null) {
          tInterface.setImplementationRef(QName.valueOf(interfaceModel.getPortTypeQName()));
        }
        context.getDefinitions().getRootElements().add(context.getBpmnObjectFactory().createInterface(tInterface));
      }
      // Add to participant
      final QName interfaceQName = new QName(context.getDefinitions().getTargetNamespace(), tInterface.getId());
      if (!participant.getInterfaceReves().contains(interfaceQName)) {
        participant.getInterfaceReves().add(interfaceQName);
      }
      // Create operations
      for (final OperationModel operationModel : interfaceModel.getOperations().values()) {
        // Find or create operation
        final String operationName = operationModel.getName();
        Operation operation = null;
        for (final Operation interfaceOperation : tInterface.getOperations()) {
          if (interfaceOperation.getName().equals(operationName)) {
            operation = interfaceOperation;
          }
        }
        if (operation == null) {
          operation = new Operation();
          operation.setId(operationModel.getId());
          operation.setName(operationName);
          tInterface.getOperations().add(operation);
        }

        // Check or add input message
        if (operation.getInMessageRef() == null) {
          WSDLMessageModel wsdlMessageModel = operationModel.getInMessageType();
          final TMessage message = generateMessage(wsdlMessageModel, context);
          operation.setInMessageRef(new QName(context.getDefinitions().getTargetNamespace(), message.getId()));
        }
        if (operationModel.getOutMessageType() != null) {
          // Check or add output message
          if (operation.getOutMessageRef() == null) {
            WSDLMessageModel wsdlMessageModel = operationModel.getOutMessageType();
            final TMessage message = generateMessage(wsdlMessageModel, context);
            operation.setOutMessageRef(new QName(context.getDefinitions().getTargetNamespace(), message.getId()));
          }
        }
      }
    }
  }

  public static TMessage generateMessage(final WSDLMessageModel wsdlMessageModel, final Designer2BpmnContext context) {
    if (!context.getMessagesMap().containsKey(wsdlMessageModel)) {
      ElementTransformer.modelToBpmnItemDefinition(wsdlMessageModel, context);
      final TItemDefinition itemDefinition = context.getDataTypeMap().get(wsdlMessageModel);
      final TMessage message = new TMessage();
      message.setName(wsdlMessageModel.getQName().getLocalPart());
      message.setId("msg" + wsdlMessageModel.getId());
      message.setItemRef(new QName(context.getDefinitions().getTargetNamespace(), itemDefinition.getId()));

      context.getDefinitions().getRootElements().add(context.getBpmnObjectFactory().createMessage(message));
      context.getMessagesMap().put(wsdlMessageModel, message);
    }
    return context.getMessagesMap().get(wsdlMessageModel);
  }

}
