/**
 * Bull SAS / OW2 Orchestra
 * Copyright (C) 2010 Bull S.A.S, and individual contributors as indicated
 * by the @authors tag.
 *
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.designer.models.transformers.designer2bpmn;

import java.util.ArrayList;
import java.util.List;
import javax.xml.namespace.QName;
import org.ow2.orchestra.designer.bpmn.model.AbstractElementWithPosition;
import org.ow2.orchestra.designer.bpmn.model.process.AbstractBPMNElementWithPosition;
import org.ow2.orchestra.designer.bpmn.model.process.SequenceFlowModel;
import org.ow2.orchestra.designer.models.exceptions.TransformationException;
import org.ow2.orchestra.jaxb.bpmn.DataInput;
import org.ow2.orchestra.jaxb.bpmn.DataInputAssociation;
import org.ow2.orchestra.jaxb.bpmn.TExclusiveGateway;
import org.ow2.orchestra.jaxb.bpmn.TExpression;
import org.ow2.orchestra.jaxb.bpmn.TFlowElement;
import org.ow2.orchestra.jaxb.bpmn.TSequenceFlow;
import org.ow2.orchestra.jaxb.bpmn.bpmndi.BPMNEdge;
import org.ow2.orchestra.jaxb.bpmn.dc.Point;
import org.ow2.orchestra.jaxb.bpmn.di.DiagramElement;
import org.ow2.orchestra.jaxb.bpmn.di.Shape;

/**
 * Transforms {@link SequenceFlowModel}s into {@link TSequenceFlow}s and vice
 * versa
 *
 * @author Loic Albertin
 */
public final class SequenceFlowTransformer {

  /**
   * Utility classes should not have public or default constructors.
   */
  private SequenceFlowTransformer() {

  }

  /**
   * @param seqFlowModel
   * @param context
   * @return
   * @throws TransformationException
   */
  public static void modelToBpmn(final SequenceFlowModel seqFlowModel, final Designer2BpmnContext context)
          throws TransformationException {
    final TSequenceFlow tSequenceFlow = new TSequenceFlow();
    tSequenceFlow.setId(seqFlowModel.getId());

    // Check source and target existence
    final AbstractElementWithPosition source = seqFlowModel.getSourceElement();
    if (source == null) {
      throw new TransformationException("BPMN conversion error: sequence flow should have a source element.");
    }

    final AbstractElementWithPosition target = seqFlowModel.getTargetElement();
    if (target == null) {
      throw new TransformationException("BPMN conversion error: sequence flow should have a target element.");
    }
    // Now it's safe to register elements in lists
    // Add to list before calling ElementTransformer.modelToBpmn (can cause recursion and duplicate elements)
    context.getModelsMap().put(seqFlowModel, tSequenceFlow);

    if (!context.getModelsMap().containsKey(source)) {
      ElementTransformer.modelToBpmn(source, context);
    }
    // Retrieve the corresponding source element
    final TFlowElement flowElementSource = context.getModelsMap().get(seqFlowModel.getSourceElement());
    tSequenceFlow.setSourceRef(flowElementSource);

    if (!context.getModelsMap().containsKey(target)) {
      ElementTransformer.modelToBpmn(target, context);
    }
    // Retrieve the corresponding target element
    final TFlowElement flowElementTarget = context.getModelsMap().get(seqFlowModel.getTargetElement());
    tSequenceFlow.setTargetRef(flowElementTarget);

    // Add condition
    if (seqFlowModel.getHasCondition()) {
      final List<DataInput> dataInputs = new ArrayList<DataInput>();
      DataInput dataInput = new DataInput();
      dataInput.setName("condition");
      dataInput.setId("condition");
      dataInputs.add(dataInput);

      final List<DataInputAssociation> dataInputAssociations =
              DataMapperTransformer.modelToBPMN(seqFlowModel.getDataMappingModel(),
                      dataInputs,
                      ((AbstractBPMNElementWithPosition) seqFlowModel.getSourceElement()).getProcessModel().getDataObjects(),
                      context);
      final TExpression conditionExpression = dataInputAssociations.get(0).getAssignments().get(0).getFrom();
      tSequenceFlow.setConditionExpression(conditionExpression);
    }

    // Set default sequenceFlow
    if (seqFlowModel.getIsDefault()) {
      if (!(flowElementSource instanceof TExclusiveGateway)) {
        throw new TransformationException("BPMN conversion error: default sequence flow should have an exclusive gateway as source element.");
      }
      final TExclusiveGateway sourceExclusiveGateway = (TExclusiveGateway) flowElementSource;
      if (sourceExclusiveGateway.getDefault() != null) {
        throw new TransformationException("BPMN conversion error: an exclusive gateway can have only one default sequence flow.");
      }
      sourceExclusiveGateway.setDefault(tSequenceFlow);
    }

    context.getTProcess().getFlowElements().add(context.getBpmnObjectFactory().createSequenceFlow(tSequenceFlow));

    // Now deals with the graphical part
    final BPMNEdge edge = new BPMNEdge();
    edge.setId("edge" + seqFlowModel.getId());
    edge.setBpmnElement(new QName(context.getDefinitions().getTargetNamespace(), tSequenceFlow.getId()));

    // The source or target definition should only be present if the edge is
    // depicted between a different source or target than the one referenced by
    // the BPMN model element of the BPMNEdge. Only the different source or
    // target is REQUIRED. Both attributes should be present only if both are
    // different. This is the case, for instance, if a Message Flow target is
    // not depicted in the current diagram because it is inside a black box
    // Pool.

    // edge.setSourceElement(new
    // QName(context.getDefinitions().getTargetNamespace(),
    // flowElementSource.getId()));
    // edge.setTargetElement(new
    // QName(context.getDefinitions().getTargetNamespace(),
    // flowElementTarget.getId()));

    // Currently sequence flow way points are not modeled in the designer model
    // so we will just get bounds values of corresponding shapes.
    // Lets retrieve shapes
    final Point targetPt = new Point();
    targetPt.setX(0);
    targetPt.setY(0);
    final DiagramElement targetDiagElem = context.getDiagramElementOnPlane(new QName(context.getDefinitions().getTargetNamespace(),
            flowElementTarget.getId()));
    if (targetDiagElem instanceof Shape) {
      targetPt.setX(((Shape) targetDiagElem).getBounds().getX());
      targetPt.setY(((Shape) targetDiagElem).getBounds().getY());
    }

    final Point sourcePt = new Point();
    sourcePt.setX(0);
    sourcePt.setY(0);
    final DiagramElement sourceDiagElem = context.getDiagramElementOnPlane(new QName(context.getDefinitions().getTargetNamespace(),
            flowElementSource.getId()));
    if (sourceDiagElem instanceof Shape) {
      sourcePt.setX(((Shape) sourceDiagElem).getBounds().getX());
      sourcePt.setY(((Shape) sourceDiagElem).getBounds().getY());
    }

    edge.getWaypoints().add(sourcePt);
    edge.getWaypoints().add(targetPt);

    context.getBpmnPlane().getDiagramElements().add(context.getBpmndiObjectFactory().createBPMNEdge(edge));
  }

}
