/**
 * Bull SAS / OW2 Orchestra
 * Copyright (C) 2010 Bull S.A.S, and individual contributors as indicated
 * by the @authors tag.
 *
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.designer.models.transformers.designer2bpmn;

import java.util.List;
import javax.xml.bind.JAXBElement;
import javax.xml.namespace.QName;
import org.ow2.orchestra.designer.bpmn.model.data.DataInputModel;
import org.ow2.orchestra.designer.bpmn.model.data.DataObjectModel;
import org.ow2.orchestra.designer.bpmn.model.data.type.WSDLMessageModel;
import org.ow2.orchestra.designer.bpmn.model.operations.CorrelationMappingModel;
import org.ow2.orchestra.designer.bpmn.model.process.TaskModel;
import org.ow2.orchestra.designer.bpmn.model.process.TaskType;
import org.ow2.orchestra.designer.models.exceptions.TransformationException;
import org.ow2.orchestra.jaxb.bpmn.DataInput;
import org.ow2.orchestra.jaxb.bpmn.DataInputAssociation;
import org.ow2.orchestra.jaxb.bpmn.DataOutput;
import org.ow2.orchestra.jaxb.bpmn.DataOutputAssociation;
import org.ow2.orchestra.jaxb.bpmn.InputSet;
import org.ow2.orchestra.jaxb.bpmn.IoSpecification;
import org.ow2.orchestra.jaxb.bpmn.MessageFlow;
import org.ow2.orchestra.jaxb.bpmn.Operation;
import org.ow2.orchestra.jaxb.bpmn.OutputSet;
import org.ow2.orchestra.jaxb.bpmn.Participant;
import org.ow2.orchestra.jaxb.bpmn.ResourceAssignmentExpression;
import org.ow2.orchestra.jaxb.bpmn.TCollaboration;
import org.ow2.orchestra.jaxb.bpmn.TDataObject;
import org.ow2.orchestra.jaxb.bpmn.TFormalExpression;
import org.ow2.orchestra.jaxb.bpmn.TInterface;
import org.ow2.orchestra.jaxb.bpmn.TMessage;
import org.ow2.orchestra.jaxb.bpmn.TPotentialOwner;
import org.ow2.orchestra.jaxb.bpmn.TRootElement;
import org.ow2.orchestra.jaxb.bpmn.TServiceTask;
import org.ow2.orchestra.jaxb.bpmn.TTask;
import org.ow2.orchestra.jaxb.bpmn.TUserTask;

/**
 * Transform {@link TaskModel}s into {@link TTask}s
 *
 * @author Guillaume Porcher
 */
public final class TaskTransformer {

  /**
   * Utility classes should not have public or default constructors.
   */
  private TaskTransformer() {

  }


  public static void modelToBpmn(final TaskModel taskModel, final Designer2BpmnContext context) throws TransformationException {
    if (taskModel.getId() == null || taskModel.getId().equals("")) {
      throw new TransformationException("BPMN conversion error: A task has no id.");
    }
    final TTask tTask;
    final JAXBElement<? extends TTask> jaxbElement;

    if (taskModel.getTaskType().equals(TaskType.SERVICE)) {
      tTask = new TServiceTask();
      jaxbElement = context.getBpmnObjectFactory().createServiceTask((TServiceTask) tTask);

      // Add or get operation
      if (taskModel.getOperation() != null) {
        TMessage inMessage = null;
        TMessage outMessage = null;
        final Operation operation = TaskTransformer.modelToBpmnOperation(taskModel, context);
        ((TServiceTask) tTask).setOperationRef(new QName(context.getDefinitions().getTargetNamespace(), operation.getId()));
        // get associated messages
        for (final JAXBElement<? extends TRootElement> rootElement : context.getDefinitions().getRootElements()) {
          if (rootElement.getValue() instanceof TMessage) {
            if (((TMessage) rootElement.getValue()).getId().equals(operation.getInMessageRef().getLocalPart())) {
              inMessage = (TMessage) rootElement.getValue();
            }
            if (operation.getOutMessageRef() != null &&
                ((TMessage) rootElement.getValue()).getId().equals(operation.getOutMessageRef().getLocalPart())) {
              outMessage = (TMessage) rootElement.getValue();
            }
          }
        }

        // Add message flows
        QName taskQName = new QName(context.getDefinitions().getTargetNamespace(), taskModel.getId());
        QName poolQName =
                new QName(context.getDefinitions().getTargetNamespace(),
                        taskModel.getOperation().getInterfaceModel().getPoolModel().getId());
        MessageFlow inMessageFlow = new MessageFlow();
        inMessageFlow.setId("inMsgFlow" + taskModel.getId());
        inMessageFlow.setMessageRef(operation.getInMessageRef());
        inMessageFlow.setSourceRef(taskQName);
        inMessageFlow.setTargetRef(poolQName);
        context.getCollaboration().getMessageFlows().add(inMessageFlow);

        // Add correlation
        for (CorrelationMappingModel correlationMappingModel : taskModel.getRequestCorrelationMappingModels()) {
          CorrelationTransformer.correlationMappingToBPMN(correlationMappingModel, inMessageFlow, context);
        }

        if (outMessage != null) {
          MessageFlow outMessageFlow = new MessageFlow();
          outMessageFlow.setId("outMsgFlow" + taskModel.getId());
          outMessageFlow.setMessageRef(operation.getOutMessageRef());
          outMessageFlow.setSourceRef(poolQName);
          outMessageFlow.setTargetRef(taskQName);
          context.getCollaboration().getMessageFlows().add(outMessageFlow);
          // Add correlation
          for (CorrelationMappingModel correlationMappingModel : taskModel.getResponseCorrelationMappingModels()) {
            CorrelationTransformer.correlationMappingToBPMN(correlationMappingModel, outMessageFlow, context);
          }
        }

        // Add data input
        generateInputOutput(taskModel, context, tTask,
            outMessage != null,
            inMessage.getItemRef(),
            outMessage != null ? outMessage.getItemRef() : null);
      }
    } else if (taskModel.getTaskType().equals(TaskType.USER)) {
      tTask = new TUserTask();
      jaxbElement = context.getBpmnObjectFactory().createUserTask((TUserTask) tTask);
      // Add or get operation
      if (taskModel.getOperation() != null) {
        TMessage inMessage = null;
        TMessage outMessage = null;
        final Operation operation = TaskTransformer.modelToBpmnOperation(taskModel, context);
        // get associated messages
        for (final JAXBElement<? extends TRootElement> rootElement : context.getDefinitions().getRootElements()) {
          if (rootElement.getValue() instanceof TMessage) {
            if (((TMessage) rootElement.getValue()).getId().equals(operation.getInMessageRef().getLocalPart())) {
              inMessage = (TMessage) rootElement.getValue();
            }
            if (operation.getOutMessageRef() != null &&
                ((TMessage) rootElement.getValue()).getId().equals(operation.getOutMessageRef().getLocalPart())) {
              outMessage = (TMessage) rootElement.getValue();
            }
          }
        }

        // Add data input
        generateInputOutput(taskModel, context, tTask,
                outMessage != null,
                inMessage.getItemRef(),
                outMessage != null ? outMessage.getItemRef() : null);
      }
      TPotentialOwner potentialOwner = new TPotentialOwner();
      ResourceAssignmentExpression potentialOwnerAssignmentExpression = new ResourceAssignmentExpression();
      TFormalExpression potentialOwnerExpression = new TFormalExpression();
      potentialOwnerExpression.getContent().add(taskModel.getPotentialOwnerExpression());
      potentialOwnerAssignmentExpression.setExpression(context.getBpmnObjectFactory().createFormalExpression(potentialOwnerExpression));
      potentialOwner.setResourceAssignmentExpression(potentialOwnerAssignmentExpression);

      tTask.getResourceRoles().add(context.getBpmnObjectFactory().createPotentialOwner(potentialOwner));
    } else {
      tTask = new TTask();
      jaxbElement = context.getBpmnObjectFactory().createTask(tTask);
    }
    tTask.setId(taskModel.getId());
    tTask.setName(taskModel.getName());


    context.getModelsMap().put(taskModel, tTask);

    context.getTProcess().getFlowElements().add(jaxbElement);

    // Now deals with positions
    ShapeTransformer.modelToBpmn(taskModel, context);

  }

  private static void generateInputOutput(TaskModel taskModel, Designer2BpmnContext context, TTask tTask, boolean hasOutput, QName inputType, QName outputType) throws TransformationException {
    // Add data input
    IoSpecification ioSpecification = new IoSpecification();
    tTask.setIoSpecification(ioSpecification);
    InputSet inputSet = new InputSet();
    for (DataInputModel dataInputModel : taskModel.getDataInputs()) {
      final DataInput dataInput = new DataInput();
      ElementTransformer.modelToBpmnItemDefinition(dataInputModel.getType(), context);
      dataInput.setItemSubjectRef(
              new QName(
                      context.getDefinitions().getTargetNamespace(),
                      dataInputModel.getType().getId()));
      dataInput.setName(dataInputModel.getName());
      dataInput.setId(dataInputModel.getId());

      ioSpecification.getDataInputs().add(dataInput);
      inputSet.getDataInputRefs().add(context.getBpmnObjectFactory().createInputSetDataInputRefs(dataInput));
    }
    ioSpecification.getInputSets().add(inputSet);

    final List<DataInputAssociation> dataInputAssociations =
            DataMapperTransformer.modelToBPMN(
                    taskModel.getDataMappingModel(),
                    tTask.getIoSpecification().getDataInputs(),
                    taskModel.getProcessModel().getDataObjects(),
                    context);
    tTask.getDataInputAssociations().addAll(dataInputAssociations);

    // Add data output
    if (hasOutput) {
      List<DataObjectModel> variables = taskModel.getOutputVariables();
      if (variables.isEmpty()) {
        throw new TransformationException("BPMN conversion error: No variable defined for task " + taskModel.getName() + ".");
      }
      for (DataObjectModel variable : variables) {
        // create dataObject in processTransformer
        ElementTransformer.modelToBpmnDataObject(context, variable);
        TDataObject dataObject = (TDataObject) context.getModelsMap().get(variable);
        // Add data output
        final DataOutput dataOutput = new DataOutput();
        ElementTransformer.modelToBpmnItemDefinition(variable.getType(), context);
        dataOutput.setItemSubjectRef(
                new QName(
                        context.getDefinitions().getTargetNamespace(),
                        variable.getType().getId()));
        if (variables.size() == 1) {
          WSDLMessageModel wsdlMessageModel = (WSDLMessageModel) taskModel.getOperation().getOutMessageType();
          dataOutput.setName(wsdlMessageModel.getParts().get(0).getPartName());
        } else {
          int index = variable.getName().lastIndexOf(" ");
          dataOutput.setName(variable.getName().substring(index));
        }
        dataOutput.setId("dataOutput" + dataObject.getId());
        ioSpecification.getDataOutputs().add(dataOutput);
        OutputSet outputSet = new OutputSet();
        outputSet.getDataOutputRefs().add(context.getBpmnObjectFactory().createOutputSetDataOutputRefs(dataOutput));
        ioSpecification.getOutputSets().add(outputSet);

        final DataOutputAssociation dataOutputAssociation = new DataOutputAssociation();
        dataOutputAssociation.getSourceReves().add(context.getBpmnObjectFactory().createTDataAssociationSourceRef(dataOutput));
        dataOutputAssociation.setTargetRef(dataObject);

        tTask.getDataOutputAssociations().add(dataOutputAssociation);
      }
    } else {
      // add empty outputSet to ioSpecification
      ioSpecification.getOutputSets().add(new OutputSet());
    }
  }

  /**
   * Create interface and operation for a task.
   *
   * @param taskModel
   * @param context
   * @return
   * @throws TransformationException
   */
  private static Operation modelToBpmnOperation(final TaskModel taskModel, final Designer2BpmnContext context) throws TransformationException {

    final String interfaceName = taskModel.getOperation().getInterfaceModel().getName();
    // Find interface
    TInterface tInterface = null;
    for (final JAXBElement<? extends TRootElement> rootElement : context.getDefinitions().getRootElements()) {
      if (rootElement.getValue() instanceof TInterface && ((TInterface) rootElement.getValue()).getName().equals(interfaceName)) {
        tInterface = (TInterface) rootElement.getValue();
      }
    }
    if (tInterface == null) {
      throw new TransformationException("BPMN conversion error: task " + taskModel.getName() + " use an unknown interface " + interfaceName);
    }

    // Check interface is in collaboration
    if (context.getCollaboration() == null) {
      throw new TransformationException("BPMN conversion error: no collaboration defined for task " + taskModel.getName());
    }
    final TCollaboration collaboration = context.getCollaboration();
    Participant interfaceParticipant = null;
    final QName interfaceQName = new QName(context.getDefinitions().getTargetNamespace(), tInterface.getId());
    for (final Participant participant : collaboration.getParticipants()) {
      if (participant.getInterfaceReves().contains(interfaceQName)) {
        interfaceParticipant = participant;
        break;
      }
    }
    if (interfaceParticipant == null) {
      throw new TransformationException("BPMN conversion error: no participant found for task " + taskModel.getName());
    }
    // Find or create operation
    final String operationName = taskModel.getOperation().getName();
    Operation operation = null;
    for (final Operation interfaceOperation : tInterface.getOperations()) {
      if (interfaceOperation.getName().equals(operationName)) {
        operation = interfaceOperation;
      }
    }
    if (operation == null) {
      throw new TransformationException("BPMN conversion error: no operation " + operationName + " defined for interface " + interfaceName + " (task " + taskModel.getName() + ")");
    }
    return operation;
  }
}
