/**
 * Copyright (C) 2010  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.designer.models.transformers.designer2bpmn;

import com.ibm.wsdl.extensions.schema.SchemaConstants;
import com.ibm.wsdl.extensions.schema.SchemaImpl;
import com.ibm.wsdl.extensions.soap.SOAPAddressImpl;
import com.ibm.wsdl.extensions.soap.SOAPBindingImpl;
import com.ibm.wsdl.extensions.soap.SOAPBodyImpl;
import com.ibm.wsdl.extensions.soap.SOAPOperationImpl;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.IOException;
import java.util.Map.Entry;
import javax.wsdl.Binding;
import javax.wsdl.BindingInput;
import javax.wsdl.BindingOperation;
import javax.wsdl.BindingOutput;
import javax.wsdl.Definition;
import javax.wsdl.Input;
import javax.wsdl.Message;
import javax.wsdl.Output;
import javax.wsdl.Part;
import javax.wsdl.Port;
import javax.wsdl.PortType;
import javax.wsdl.Service;
import javax.wsdl.WSDLException;
import javax.wsdl.extensions.schema.Schema;
import javax.wsdl.extensions.soap.SOAPAddress;
import javax.wsdl.extensions.soap.SOAPBinding;
import javax.wsdl.extensions.soap.SOAPBody;
import javax.wsdl.extensions.soap.SOAPOperation;
import javax.wsdl.factory.WSDLFactory;
import javax.wsdl.xml.WSDLWriter;
import javax.xml.XMLConstants;
import javax.xml.namespace.QName;
import org.apache.ws.commons.schema.XmlSchema;
import org.apache.ws.commons.schema.XmlSchemaCollection;
import org.apache.ws.commons.schema.XmlSchemaComplexType;
import org.apache.ws.commons.schema.XmlSchemaElement;
import org.apache.ws.commons.schema.XmlSchemaForm;
import org.apache.ws.commons.schema.XmlSchemaObject;
import org.apache.ws.commons.schema.XmlSchemaSequence;
import org.apache.ws.commons.schema.XmlSchemaSerializer.XmlSchemaSerializerException;
import org.apache.ws.commons.schema.XmlSchemaSimpleType;
import org.apache.ws.commons.schema.XmlSchemaType;
import org.ow2.orchestra.designer.bpmn.model.data.type.DataTypeModel;
import org.ow2.orchestra.designer.bpmn.model.data.type.XSDSequenceType;
import org.ow2.orchestra.designer.bpmn.model.data.type.XSDSimpleType;
import org.ow2.orchestra.designer.bpmn.model.data.type.WSDLMessageModel;
import org.ow2.orchestra.designer.bpmn.model.data.type.WSDLMessagePartModel;
import org.ow2.orchestra.designer.bpmn.model.data.type.XSDElementModel;
import org.ow2.orchestra.designer.bpmn.model.data.type.XSDTypeModel;
import org.ow2.orchestra.designer.bpmn.model.operations.InterfaceModel;
import org.ow2.orchestra.designer.bpmn.model.operations.OperationModel;
import org.ow2.orchestra.designer.bpmn.model.process.AbstractBPMNElementWithPosition;
import org.ow2.orchestra.designer.bpmn.model.process.ProcessModel;
import org.ow2.orchestra.designer.bpmn.model.process.TaskModel;
import org.ow2.orchestra.util.Misc;
import org.ow2.orchestra.util.XmlUtil;


/**
 * @author Guillaume Porcher
 */
public final class WSDLGenerator {

  /**
   * Orchestra constant: replaced in orchestra engine with the server url.
   */
  private static final String ORCHESTRA_WS_URL = "${orchestra.ws.url}/";


  private WSDLGenerator() {
  }

  /**
   * wsdl ns
   */
  public static final String XMLNS_WSDL_SOAP = "http://schemas.xmlsoap.org/wsdl/soap/";

  private static final WSDLFactory WSDL_FACTORY;

  static {
    try {
      WSDL_FACTORY = WSDLFactory.newInstance();
    } catch (final WSDLException e) {
      throw new RuntimeException(e);
    }
  }

  private static Definition generateWsdlDefinition(
          final ProcessModel processModel) throws WSDLException, XmlSchemaSerializerException {
    final XmlSchemaCollection schemaCollection = new XmlSchemaCollection();
    final Definition wsdlDefinition = WSDLGenerator.WSDL_FACTORY.newDefinition();
    wsdlDefinition.setQName(new QName(processModel.getWsdlNamespace(), processModel.getName(), "tns"));
    wsdlDefinition.setTargetNamespace(processModel.getWsdlNamespace());
    wsdlDefinition.addNamespace("tns", processModel.getWsdlNamespace());
    wsdlDefinition.addNamespace("xsd", XMLConstants.W3C_XML_SCHEMA_NS_URI);
    wsdlDefinition.addNamespace("soap", WSDLGenerator.XMLNS_WSDL_SOAP);
    wsdlDefinition.setTypes(wsdlDefinition.createTypes());

    if (!processModel.getInterfaces().isEmpty()) {
      // generate WSDL port types for the process
      WSDLGenerator.generateProcessWSDL(processModel, wsdlDefinition, schemaCollection);
    }
    // Add types definition
    wsdlDefinition.setTypes(wsdlDefinition.createTypes());
    for (XmlSchema schema: schemaCollection.getXmlSchemas()) {
      if (!schema.getTargetNamespace().equals(XMLConstants.W3C_XML_SCHEMA_NS_URI)) {
        Schema wsdlSchema = new SchemaImpl();
        wsdlSchema.setElement(schema.getSchemaDocument().getDocumentElement());
        wsdlSchema.setElementType(SchemaConstants.Q_ELEM_XSD_2001);
        wsdlDefinition.getTypes().addExtensibilityElement(wsdlSchema);
      }
    }
    return wsdlDefinition;
  }

  private static void generateProcessWSDL(final ProcessModel processModel, final Definition wsdlDefinition, final XmlSchemaCollection schemaCollection) {
    // add wsdl informations
    for (final InterfaceModel interfaceModel : processModel.getInterfaces().values()) {

      // generate PortType
      final PortType portType = wsdlDefinition.createPortType();
      portType.setQName(new QName(processModel.getWsdlNamespace(), interfaceModel.getName(), "tns"));
      portType.setUndefined(false);

      wsdlDefinition.addPortType(portType);

      // add soap binding for the PortType
      final Binding binding = wsdlDefinition.createBinding();
      binding.setQName(new QName(processModel.getWsdlNamespace(), interfaceModel.getName() + "SOAPBinding", "tns"));
      binding.setUndefined(false);
      binding.setPortType(portType);
      final SOAPBinding soapBinding = new SOAPBindingImpl();
      soapBinding.setStyle("document");
      soapBinding.setTransportURI("http://schemas.xmlsoap.org/soap/http");
      binding.addExtensibilityElement(soapBinding);

      for (final OperationModel op : interfaceModel.getOperations().values()) {
        // add operations to portType
        final javax.wsdl.Operation operation = wsdlDefinition.createOperation();
        operation.setName(op.getName());
        operation.setUndefined(false);

        portType.addOperation(operation);

        final WSDLMessageModel inMessageModel = op.getInMessageType();
        // generate inMessage
        final Message message = WSDLGenerator.generateWSDLMessage(processModel, inMessageModel, wsdlDefinition, schemaCollection);

        final Input input = wsdlDefinition.createInput();
        input.setMessage(message);
        operation.setInput(input);
        // generate outMessage
        if (op.getOutMessageType() != null) {
          final WSDLMessageModel outMessageModel = op.getOutMessageType();
          final Message outWsdlMessage = WSDLGenerator.generateWSDLMessage(processModel, outMessageModel, wsdlDefinition, schemaCollection);

          final Output output = wsdlDefinition.createOutput();
          output.setMessage(outWsdlMessage);
          operation.setOutput(output);
        }


        boolean generateBinding = true;
        for (AbstractBPMNElementWithPosition receiveElement : op.getReceiveElements()) {
          if (receiveElement instanceof TaskModel) {
            // don't generate binding for task operations.
            generateBinding = false;
          }
        }
        // add binding for operation
        if (generateBinding) {
          final BindingOperation bindingOperation = wsdlDefinition.createBindingOperation();
          bindingOperation.setName(op.getName());
          bindingOperation.setOperation(operation);
          binding.addBindingOperation(bindingOperation);
          final SOAPOperation soapOperation = new SOAPOperationImpl();
          soapOperation.setSoapActionURI(processModel.getWsdlNamespace() + processModel.getName() + "/" + op.getName());
          bindingOperation.addExtensibilityElement(soapOperation);

          final BindingInput bindingInput = wsdlDefinition.createBindingInput();
          final SOAPBody inputSoapBody = new SOAPBodyImpl();
          inputSoapBody.setUse("literal");
          bindingInput.addExtensibilityElement(inputSoapBody);
          bindingOperation.setBindingInput(bindingInput);

          // generate outMessage
          if (op.getOutMessageType() != null) {
            final BindingOutput bindingOutput = wsdlDefinition.createBindingOutput();
            final SOAPBody outputSoapBody = new SOAPBodyImpl();
            outputSoapBody.setUse("literal");
            bindingOutput.addExtensibilityElement(outputSoapBody);
            bindingOperation.setBindingOutput(bindingOutput);
          }
        }
      }
      if (!binding.getBindingOperations().isEmpty()) {
      wsdlDefinition.addBinding(binding);
      }

    }
    // add service and ports
    final Service service = wsdlDefinition.createService();
    service.setQName(new QName(processModel.getWsdlNamespace(), processModel.getName() + "Service", "tns"));
    wsdlDefinition.addService(service);

    for (final Object bindingEntry : wsdlDefinition.getBindings().entrySet()) {
      final Binding binding = ((Entry<QName, Binding>) bindingEntry).getValue();
      final Port port = wsdlDefinition.createPort();
      port.setBinding(binding);
      port.setName(binding.getPortType().getQName().getLocalPart() + "Port");

      service.addPort(port);

      final SOAPAddress soapAddress = new SOAPAddressImpl();
      port.addExtensibilityElement(soapAddress);
      soapAddress.setLocationURI(WSDLGenerator.ORCHESTRA_WS_URL + processModel.getName() + "/" + port.getName());
    }
  }

  private static Message generateWSDLMessage(final ProcessModel processModel,
                                             final WSDLMessageModel messageModel,
                                             final Definition wsdlDefinition, XmlSchemaCollection schemaCollection) {
    final Message message = wsdlDefinition.createMessage();
    message.setUndefined(false);
    message.setQName(messageModel.getQName());

    for (WSDLMessagePartModel messagePartModel : messageModel.getParts()) {
      final Part messagePart = wsdlDefinition.createPart();
      // Process will use Document/literal binding
      // According to WS-Basic profile 1.1, messages contains only one part.
      messagePart.setName(messagePartModel.getPartName());

      if (messagePartModel.getXsdType() != null) {
        messagePart.setTypeName(messagePartModel.getXsdType().getQName());
        if (!messagePart.getTypeName().getNamespaceURI().equals(XMLConstants.W3C_XML_SCHEMA_NS_URI)) {
          generateXSDDefinition(schemaCollection, messagePartModel.getXsdType(), true);
        }
      } else {
        QName elementQName = messagePartModel.getXsdElement().getQName();
        messagePart.setElementName(elementQName);
        generateXSDDefinition(schemaCollection, messagePartModel.getXsdElement(), true);

      }
      message.addPart(messagePart);
    }
    wsdlDefinition.addMessage(message);
    return message;
  }

  private static XmlSchemaObject generateXSDDefinition(final XmlSchemaCollection schemaCollection, final DataTypeModel dataTypeModel, boolean topLevel) {
    QName dataTypeQName = dataTypeModel.getQName();
    if (dataTypeModel instanceof XSDElementModel) {
      // generate element xsd type
      if (schemaCollection.getElementByQName(dataTypeQName) == null) {
        XmlSchema schema = schemaCollection.schemaForNamespace(dataTypeQName.getNamespaceURI());
        if (schema == null) {
          schema = new XmlSchema(dataTypeQName.getNamespaceURI(), schemaCollection);
          schema.setElementFormDefault(XmlSchemaForm.QUALIFIED);
        }
        XmlSchemaElement schemaElement = new XmlSchemaElement(schema, topLevel);
        schemaElement.setName(dataTypeQName.getLocalPart());
        XSDTypeModel elementTypeModel = ((XSDElementModel)dataTypeModel).getType();
        QName elementTypeQName = elementTypeModel.getQName();
        if (elementTypeQName.getNamespaceURI().equals(XMLConstants.W3C_XML_SCHEMA_NS_URI)) {
          schemaElement.setSchemaTypeName(new QName(XMLConstants.W3C_XML_SCHEMA_NS_URI, elementTypeQName.getLocalPart(), "xsd"));
        } else {
          schemaElement.setSchemaType((XmlSchemaType) generateXSDDefinition(schemaCollection, elementTypeModel, false));
        }
        return schemaElement;
      }
      return schemaCollection.getElementByQName(dataTypeQName);
    } else if (dataTypeModel instanceof XSDSimpleType) {
      // generate xsd type
      if (schemaCollection.getTypeByQName(dataTypeQName) == null) {
        XmlSchema schema = schemaCollection.schemaForNamespace(dataTypeQName.getNamespaceURI());
        if (schema == null) {
          schema = new XmlSchema(dataTypeQName.getNamespaceURI(), schemaCollection);
          schema.setElementFormDefault(XmlSchemaForm.QUALIFIED);
        }
        XmlSchemaSimpleType schemaSimpleType = new XmlSchemaSimpleType(schema, topLevel);
        schemaSimpleType.setName(dataTypeQName.getLocalPart());
        return schemaSimpleType;
      }
      return schemaCollection.getTypeByQName(dataTypeQName);
    } else if (dataTypeModel instanceof XSDSequenceType) {
      // generate xsd type
      if (schemaCollection.getTypeByQName(dataTypeQName) == null) {
        XmlSchema schema = schemaCollection.schemaForNamespace(dataTypeQName.getNamespaceURI());
        if (schema == null) {
          schema = new XmlSchema(dataTypeQName.getNamespaceURI(), schemaCollection);
          schema.setElementFormDefault(XmlSchemaForm.QUALIFIED);
        }
        XmlSchemaSequence xmlSchemaSequence = new XmlSchemaSequence();
        for (XSDElementModel sequenceElement : ((XSDSequenceType) dataTypeModel).getElements()) {
          xmlSchemaSequence.getItems().add((XmlSchemaElement) generateXSDDefinition(schemaCollection, sequenceElement, false));
        }
        XmlSchemaComplexType schemaComplexType = new XmlSchemaComplexType(schema, topLevel);
        schemaComplexType.setParticle(xmlSchemaSequence);
        schemaComplexType.setName(dataTypeQName.getLocalPart());
        return schemaComplexType;
      }
      return schemaCollection.getTypeByQName(dataTypeQName);
    }
    return null;
  }


  /**
   * Write a WSDL definition.
   *
   * @param processModel the processModel to use for the generation
   * @return the wsdl generated byte content.
   */
  public static void writeWsdl(
          final ProcessModel processModel,
          final File outputFile) throws WSDLException, IOException, XmlSchemaSerializerException {
    WSDLWriter wsdlWriter = WSDL_FACTORY.newWSDLWriter();
    Definition wsdlToWrite = generateWsdlDefinition(processModel);
    // generate xml for this definition
    final ByteArrayOutputStream wsdlOutput = new ByteArrayOutputStream();
    wsdlWriter.writeWSDL(wsdlToWrite, wsdlOutput);
    // Indent WSDL output
    final String indentedOut = XmlUtil.toString(XmlUtil.getDocumentFromString(wsdlOutput.toString(), false));
    Misc.write(indentedOut, outputFile);
  }

  public static WSDLFactory getWsdlFactory() {
    return WSDL_FACTORY;
  }
}
